@@ -1,24 +1,19 @@
-/* Copyright (C) 1992, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1992, 1995, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* echogs.c */
@@ -51,8 +46,8 @@ extern int fputc(), fputs();
 /*
  * Usage:
 	echogs [-e .extn] [-(w|a)[b][-] file] [-h] [-n]
-	  (-D | -F | -x hexstring | -[q|Q} string | -[q|Q]string | -s |
-	   -i | -r file | -R file | -X)*
+	  (-d|-D | -f|-F | -x hexstring | -(l|q|Q) string | -(l|q|Q)string |
+	   -s | -u string | -i | -r file | -R file | -X)*
 	  [-] string*
  * Echoes string(s), or the binary equivalent of hexstring(s).
  * If -w, writes to file; if -a, appends to file; if neither,
@@ -63,18 +58,20 @@ extern int fputc(), fputs();
  * -e specifies an extension to be added to the file name.
  * If -h, write the output in hex instead of literally.
  * If -n, does not append a newline to the output.
- * -D means insert the date and time.
- * -F means insert the file name passed as the argument of -w or -a.
+ * -d or -D means insert the date and time.
+ * -f or -F means insert the file name passed as the argument of -w or -a.
  * -q means write the next string literally.
- * -Q means the same as -q followed by -s.
+ * -l or -Q means the same as -q followed by -s.
  * -s means write a space.
+ * -u means convert the next string to upper case.
  * -i means read from stdin, treating each line as an argument.
  * -r means read from a named file in the same way.
- * -R means read from a named file with no interpretation
+ * -R means copy a named file with no interpretation
  *   (but convert to hex if -h is in effect).
  * -X means treat any following literals as hex rather than string data.
  * - alone means treat the rest of the line as literal data,
  *   even if the first string begins with a -.
+ * -+<letter> is equivalent to -<Letter>, i.e., it upper-cases the letter.
  * Inserts spaces automatically between the trailing strings,
  * but nowhere else; in particular,
 	echogs -q a b
@@ -162,22 +159,32 @@ main(int argc, char *argv[])
 			argp++, nargs--;
 		}
 		if ( sw == 0 && arg[0] == '-' )
-		{	sp = 0;
-			switch ( arg[1] )
+		{	char chr = arg[1];
+
+			sp = 0;
+swc:			switch ( chr )
 			{
+			case '+':		/* upper-case command */
+				++arg;
+				chr = toupper(arg[1]);
+				goto swc;
+			case 'l':		/* literal string, then -s */
+				chr = 'Q';
+				/* falls through */
 			case 'q':		/* literal string */
 			case 'Q':		/* literal string, then -s */
 				if ( arg[2] != 0 )
 				  {	(*eputs)(arg + 2, out);
-					if ( arg[1] == 'Q' )
+					if ( chr == 'Q' )
 					  (*eputc)(' ', out);
 					break;
 				  }
 				/* falls through */
 			case 'r':		/* read from a file */
-			case 'R':		/* insert file literally */
+			case 'R':
+			case 'u':		/* upper-case string */
 			case 'x':		/* hex string */
-				sw = arg[1];
+				sw = chr;
 				break;
 			case 's':		/* write a space */
 				(*eputc)(' ', out);
@@ -186,7 +193,8 @@ main(int argc, char *argv[])
 				interact = 1;
 				in = stdin;
 				break;
-			case 'D':		/* insert date/time */
+			case 'd':		/* insert date/time */
+			case 'D':
 			{	time_t t;
 				char str[26];
 				time(&t);
@@ -194,7 +202,8 @@ main(int argc, char *argv[])
 				str[24] = 0;	/* remove \n */
 				(*eputs)(str, out);
 			}	break;
-			case 'F':		/* insert file name */
+			case 'f':		/* insert file name */
+			case 'F':
 				(*eputs)(fnparam, out);
 				break;
 			case 'X':		/* treat literals as hex */
@@ -240,6 +249,12 @@ main(int argc, char *argv[])
 			}
 			fclose(in);
 			break;
+		case 'u':
+		  {	const char *up;
+			for ( up = arg; *up; up++ )
+			  (*eputc)(toupper(*up), out);
+		  }	sw = 0;
+			break;
 		case 'x':
 xx:		{	char *xp;
 			unsigned int xchr = 1;
