@@ -1,24 +1,19 @@
-/* Copyright (C) 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* gdevbbox.c */
@@ -48,7 +43,7 @@ private dev_proc_close_device(gx_forward_close_device);	/* see below */
 #ifdef TEST
 private dev_proc_output_page(bbox_output_page);
 #else
-#define bbox_output_page gx_default_output_page
+#define bbox_output_page gx_forward_output_page
 #endif
 private dev_proc_fill_rectangle(bbox_fill_rectangle);
 private dev_proc_copy_mono(bbox_copy_mono);
@@ -74,8 +69,16 @@ private dev_proc_strip_copy_rop(bbox_strip_copy_rop);
 #ifndef TEST
 private const
 #endif
+/* We initialize with a very high resolution, to prevent limitchecks */
+/* if the resolution is changed to any reasonable value. */
+/* We initialize the width and height to "infinite" values, */
+/* leaving a little room for stroke widths, rounding, etc. */
+#define max_coord (min(max_int, fixed2int(max_fixed)) - 1000)
+#define max_resolution 4000
 gx_device_bbox far_data gs_bbox_device = {
-	std_device_std_body(gx_device_bbox, 0, "bbox", 0, 0, 72, 72),
+	std_device_std_body(gx_device_bbox, 0, "bbox",
+			    max_coord, max_coord,
+			    max_resolution, max_resolution),
 	{	bbox_open_device,
 		NULL,			/* get_initial_matrix */
 		NULL,			/* sync_output */
@@ -95,7 +98,7 @@ gx_device_bbox far_data gs_bbox_device = {
 		NULL,			/* get_xfont_procs */
 		NULL,			/* get_xfont_device */
 		NULL,			/* map_rgb_alpha_color */
-		NULL,			/* get_page_device */
+		gx_page_device_get_page_device,
 		NULL,			/* get_alpha_bits */
 		bbox_copy_alpha,
 		NULL,			/* get_band */
@@ -115,6 +118,8 @@ gx_device_bbox far_data gs_bbox_device = {
 	},
 	0				/* target */
 };
+#undef max_coord
+#undef max_resolution
 
 /* Copy device parameters back from the target. */
 private void
@@ -137,14 +142,6 @@ bbox_copy_params(gx_device_bbox *bdev)
 #undef copy_param
 #undef copy_array_param
 	  }
-	else
-	  { /* If no target, make the width and height "infinite". */
-	    /* Leave some room for stroke widths, rounding, etc. */
-#define max_coord (min(max_int, fixed2int(max_fixed)) - 1000)
-	    gx_device_set_width_height((gx_device *)bdev,
-				       max_coord, max_coord);
-#undef max_coord
-	  }
 	if ( dev_proc(bdev, map_rgb_color) != 0 )
 	  bdev->white =
 	    (*dev_proc(bdev, map_rgb_color))
@@ -271,7 +268,10 @@ private int
 bbox_copy_mono(gx_device *dev, const byte *data,
   int dx, int raster, gx_bitmap_id id, int x, int y, int w, int h,
   gx_color_index zero, gx_color_index one)
-{	bbox_add_int_rect(&bdev->bbox, x, y, x + w, y + h);
+{	if ( (one != gx_no_color_index && one != bdev->white) ||
+	     (zero != gx_no_color_index && zero != bdev->white)
+	   )
+	  bbox_add_int_rect(&bdev->bbox, x, y, x + w, y + h);
 	/* gx_forward_copy_mono doesn't exist */
 	{ gx_device *tdev = bdev->target;
 	  return (tdev == 0 ? 0 :
@@ -304,7 +304,8 @@ bbox_draw_line(gx_device *dev,
 	  yp = y0, yq = y1 + 1;
 	else
 	  yp = y1, yq = y0 + 1;
-	bbox_add_int_rect(&bdev->bbox, xp, yp, xq, yq);
+	if ( color != bdev->white )
+	  bbox_add_int_rect(&bdev->bbox, xp, yp, xq, yq);
 	/* gx_forward_draw_line doesn't exist */
 	{ gx_device *tdev = bdev->target;
 	  return (tdev == 0 ? 0 :
@@ -371,19 +372,50 @@ bbox_get_params(gx_device *dev, gs_param_list *plist)
 
 	if ( code < 0 )
 	  return code;
-	bbox[0] = bdev->bbox.p.x;
-	bbox[1] = bdev->bbox.p.y;
-	bbox[2] = bdev->bbox.q.x;
-	bbox[3] = bdev->bbox.q.y;
+	bbox[0] = fixed2float(bdev->bbox.p.x);
+	bbox[1] = fixed2float(bdev->bbox.p.y);
+	bbox[2] = fixed2float(bdev->bbox.q.x);
+	bbox[3] = fixed2float(bdev->bbox.q.y);
 	bba.data = bbox, bba.size = 4, bba.persistent = false;
 	return param_write_float_array(plist, "PageBoundingBox", &bba);
 }
 
 /* We implement put_params to ensure that we keep the important */
-/* device parameters up to date. */
+/* device parameters up to date, and to prevent an /undefined error */
+/* from PageBoundingBox. */
 private int
 bbox_put_params(gx_device *dev, gs_param_list *plist)
-{	int code = gx_forward_put_params(dev, plist);
+{	int code;
+	int ecode = 0;
+	gs_param_name param_name;
+	gs_param_float_array bba;
+
+	code = param_read_float_array(plist, (param_name = "PageBoundingBox"),
+				      &bba);
+	switch ( code )
+	  {
+	  case 0:
+	    if ( bba.size != 4 )
+	      { ecode = gs_note_error(gs_error_rangecheck);
+	        goto e;
+	      }
+	    break;
+	  default:
+	    ecode = code;
+e:	    param_signal_error(plist, param_name, ecode);
+	  case 1:
+	    bba.data = 0;
+	  }
+
+	code = gx_forward_put_params(dev, plist);
+	if ( ecode < 0 )
+	  code = ecode;
+	if ( code >= 0 && bba.data != 0 )
+	  { bdev->bbox.p.x = float2fixed(bba.data[0]);
+	    bdev->bbox.p.y = float2fixed(bba.data[1]);
+	    bdev->bbox.q.x = float2fixed(bba.data[2]);
+	    bdev->bbox.q.y = float2fixed(bba.data[3]);
+	  }
 	bbox_copy_params(bdev);
 	return code;
 }
@@ -394,7 +426,7 @@ private fixed
 edge_x_at_y(const gs_fixed_edge *edge, fixed y)
 {	return fixed_mult_quo(edge->end.x - edge->start.x,
 			      y - edge->start.y,
-			      edge->end.y - edge->start.y);
+			      edge->end.y - edge->start.y) + edge->start.x;
 }
 private int
 bbox_fill_trapezoid(gx_device *dev,
@@ -462,7 +494,6 @@ bbox_fill_triangle(gx_device *dev,
 		  (*dev_proc(tdev, fill_triangle))(tdev, px, py, ax, ay,
 						   bx, by, pdevc, lop));
 	}
-
 }
 
 private int
@@ -535,8 +566,10 @@ bbox_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 
 	    if ( gx_path_bbox(ppath, &ibox) < 0 )
 	      return 0;
-	    gx_stroke_expansion(pis, &expand);
-	    adjust_box(&ibox, expand);
+	    if ( gx_stroke_path_expansion(pis, ppath, &expand) < 0 )
+	      ibox.p.x = ibox.p.y = min_fixed, ibox.q.x = ibox.q.y = max_fixed;
+	    else
+	      adjust_box(&ibox, expand);
 	    if ( pcpath != NULL &&
 		 !gx_cpath_includes_rectangle(pcpath, ibox.p.x, ibox.p.y,
 					      ibox.q.x, ibox.q.y)
@@ -596,6 +629,8 @@ typedef struct bbox_image_enum_s {
 	gs_matrix matrix;	/* map from image space to device space */
 	const gx_clip_path *pcpath;
 	void *target_info;
+	int x0, x1;
+	int y, height;
 } bbox_image_enum;
 gs_private_st_ptrs2(st_bbox_image_enum, bbox_image_enum, "bbox_image_enum",
   bbox_image_enum_enum_ptrs, bbox_image_enum_reloc_ptrs, pcpath, target_info);
@@ -603,7 +638,7 @@ gs_private_st_ptrs2(st_bbox_image_enum, bbox_image_enum, "bbox_image_enum",
 private int
 bbox_begin_image(gx_device *dev,
   const gs_imager_state *pis, const gs_image_t *pim,
-  gs_image_format_t format, gs_image_shape_t shape,
+  gs_image_format_t format, const gs_int_rect *prect,
   const gx_drawing_color *pdcolor, const gx_clip_path *pcpath,
   gs_memory_t *memory, void **pinfo)
 {	int code;
@@ -622,11 +657,19 @@ bbox_begin_image(gx_device *dev,
 	pbe->matrix = mat;
 	pbe->pcpath = pcpath;
 	pbe->target_info = 0;		/* in case no target */
+	if ( prect )
+	  { pbe->x0 = prect->p.x, pbe->x1 = prect->q.x;
+	    pbe->y = prect->p.y, pbe->height = prect->q.y - prect->p.y;
+	  }
+	else
+	  { pbe->x0 = 0, pbe->x1 = pim->Width;
+	    pbe->y = 0, pbe->height = pim->Height;
+	  }
 	*pinfo = pbe;
 	/* Skip the call if there is no target. */
 	{ gx_device *tdev = bdev->target;
 	  return (tdev == 0 ? 0 :
-		  (*dev_proc(tdev, begin_image))(tdev, pis, pim, format, shape,
+		  (*dev_proc(tdev, begin_image))(tdev, pis, pim, format, prect,
 						 pdcolor, pcpath, memory,
 						 &pbe->target_info));
 	}
@@ -634,8 +677,7 @@ bbox_begin_image(gx_device *dev,
 
 private int
 bbox_image_data(gx_device *dev,
-  void *info, const byte **planes, uint raster,
-  int x, int y, int width, int height)
+  void *info, const byte **planes, int data_x, uint raster, int height)
 {	gx_device *tdev = bdev->target;
 	bbox_image_enum *pbe = info;
 	const gx_clip_path *pcpath = pbe->pcpath;
@@ -643,10 +685,10 @@ bbox_image_data(gx_device *dev,
 	gs_point corners[4];
 	gs_fixed_rect ibox;
 
-	sbox.p.x = x;
-	sbox.p.y = y;
-	sbox.q.x = x + width;
-	sbox.q.y = y + height;
+	sbox.p.x = pbe->x0;
+	sbox.p.y = pbe->y;
+	sbox.q.x = pbe->x1;
+	sbox.q.y = pbe->y += height;
 	gs_bbox_transform_only(&sbox, &pbe->matrix, corners);
 	gs_points_bbox(corners, &dbox);
 	ibox.p.x = float2fixed(dbox.p.x);
@@ -686,18 +728,21 @@ bbox_image_data(gx_device *dev,
 	    bbox_add_rect(&bdev->bbox, ibox.p.x, ibox.p.y, ibox.q.x, ibox.q.y);
 	  }
 	/* Skip the call if there is no target. */
-	return (tdev == 0 ? 0 :
+	return (tdev == 0 ? pbe->y >= pbe->height :
 		(*dev_proc(tdev, image_data))(tdev, pbe->target_info, planes,
-					      raster, x, y, width, height));
+					      data_x, raster, height));
 }
 
 private int
 bbox_end_image(gx_device *dev, void *info, bool draw_last)
 {	bbox_image_enum *pbe = info;
+	void *target_info = pbe->target_info;
 	/* Skip the call if there is no target. */
 	gx_device *tdev = bdev->target;
+	int code =
+	  (tdev == 0 ? 0 :
+	   (*dev_proc(tdev, end_image))(tdev, target_info, draw_last));
 
-	return (tdev == 0 ? 0 :
-		(*dev_proc(tdev, end_image))(tdev, pbe->target_info,
-					     draw_last));
+	gs_free_object(pbe->memory, pbe, "bbox_end_image");
+	return code;
 }
