@@ -1,24 +1,19 @@
-/* Copyright (C) 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* gdevpdf.c */
@@ -40,6 +35,10 @@
 #include "strimpl.h"		/* for short-sighted compilers */
 #include "scfx.h"		/* s_CFE_template is default */
 
+/* Define the size of the internal stream buffer. */
+/* (This is not a limitation, it only affects performance.) */
+#define sbuf_size 512
+
 /* GC descriptors */
 private_st_pdf_resource();
 private_st_pdf_font();
@@ -105,28 +104,13 @@ gx_device_pdf far_data gs_pdfwrite_device =
 		gdev_pdf_image_data,
 		gdev_pdf_end_image
 	},
-	 {	/* General parameters */
-	   0/*false*/, 0/*false*/, 1/*true*/,
-		/* Color sampled image parameters */
-	    { 0/*false*/, 72, 1/*true*/, 0, NULL, -1, 0/*false*/ },
-	   /*,*/ 1/*true*/,
-		/* Grayscale sampled image parameters */	  
-	    { 0/*false*/, 72, 1/*true*/, 0, NULL, -1, 0/*false*/ },
-	   /*,*/
-		/* Monochrome sampled image parameters */
-	    { 0/*false*/, 72, 1/*true*/, "CCITTFaxEncode", &s_CFE_template, -1, 0/*false*/ },
-	   /*,*/
-		/* Font embedding parameters */
-	    { 0, 0, 1/*true*/ }, { 0, 0, 1/*true*/ },
-	   1/*true*/, 1/*true*/, 20
-	 },
+	psdf_initial_values(0 /*false*/),	/* (!ASCII85EncodePages) */
+	1.2,		/* CompatibilityLevel */
+	0/*false*/,	/* DoThumbnails */
+	1,		/* FirstObjectNumber */
 	0/*false*/,	/* binary_ok */
-	0,		/* pdf_memory */
-	 { 0 },		/* fname */
 	 { 0 },		/* tfname */
-	0,		/* file */
 	0,		/* tfile */
-	 { X_DPI/72.0, Y_DPI/72.0 },	/* scale */
 	0,		/* next_id */
 	0,		/* root_id */
 	0,		/* info_id */
@@ -135,17 +119,14 @@ gx_device_pdf far_data gs_pdfwrite_device =
 	0,		/* next_page */
 	 { 0 },		/* contents_ids */
 	0,		/* next_contents_id */
-	0,		/* in_contents */
+	pdf_in_none,	/* context */
 	0,		/* contents_length_id */
 	0,		/* contents_pos */
-	gx_no_color_index,	/* fill_color */
-	gx_no_color_index,	/* stroke_color */
-	gs_no_id,	/* clip_path_id */
-	0,		/* procsets */
+	NoMarks,	/* procsets */
 	-1,		/* flatness */
 	 { gx_line_params_initial },	/* line_params */
-	 { 0 },		/* dash_pattern */
-	 { 0 },		/* page_ids */
+	0,		/* page_ids */
+	0,		/* num_page_ids */
 	0,		/* pages_referenced */
 	 { 0 },		/* resources */
 	0,		/* annots */
@@ -166,15 +147,24 @@ gx_device_pdf far_data gs_pdfwrite_device =
 
 /* ------ Document ------ */
 
+/* Initialize the IDs allocated at startup. */
+void
+pdf_initialize_ids(gx_device_pdf *pdev)
+{	pdev->next_id = pdev->FirstObjectNumber;
+	pdev->root_id = pdf_obj_ref(pdev);
+	pdev->pages_id = pdf_obj_ref(pdev);
+}
+
 /* Open the document if necessary. */
 void
 pdf_open_document(gx_device_pdf *pdev)
-{	FILE *file = pdev->file;
-	if ( !in_document(pdev) && ftell(file) == 0 )
-	  { fputs("%PDF-1.1\n", file);
+{	stream *s = pdev->strm;
+
+	if ( !is_in_document(pdev) && stell(s) == 0 )
+	  { pputs(s, "%PDF-1.1\n");
 	    pdev->binary_ok = !pdev->params.ASCII85EncodePages;
 	    if ( pdev->binary_ok )
-	      fputs("%\307\354\217\242\n", file);
+	      pputs(s, "%\307\354\217\242\n");
 	  }
 }
 
@@ -190,8 +180,8 @@ pdf_next_id(gx_device_pdf *pdev)
 long
 pdf_obj_ref(gx_device_pdf *pdev)
 {	long id = pdf_next_id(pdev);
-	FILE *file = pdev->file;
-	long pos = ftell(file);
+	stream *s = pdev->strm;
+	long pos = stell(s);
 	FILE *tfile = pdev->tfile;
 
 	fwrite(&pos, sizeof(pos), 1, tfile);
@@ -201,27 +191,29 @@ pdf_obj_ref(gx_device_pdf *pdev)
 /* Begin an object, optionally allocating an ID. */
 long
 pdf_open_obj(gx_device_pdf *pdev, long id)
-{	FILE *file = pdev->file;
-	long pos = ftell(file);
+{	stream *s = pdev->strm;
 	FILE *tfile = pdev->tfile;
 
 	if ( id <= 0 )
 	  { id = pdf_obj_ref(pdev);
 	  }
 	else
-	  { long tpos = ftell(tfile);
-	    fseek(tfile, (id - 1) * sizeof(pos), SEEK_SET);
+	  { long pos = stell(s);
+	    long tpos = ftell(tfile);
+
+	    fseek(tfile, (id - pdev->FirstObjectNumber) * sizeof(pos),
+		  SEEK_SET);
 	    fwrite(&pos, sizeof(pos), 1, tfile);
 	    fseek(tfile, tpos, SEEK_SET);
 	  }
-	fprintf(file, "%ld 0 obj\n", id);
+	pprintld1(s, "%ld 0 obj\n", id);
 	return id;
 }
 
 /* End an object. */
 int
 pdf_end_obj(gx_device_pdf *pdev)
-{	fputs("endobj\n", pdev->file);
+{	pputs(pdev->strm, "endobj\n");
 	return 0;
 }
 
@@ -230,17 +222,10 @@ pdf_end_obj(gx_device_pdf *pdev)
 /* Set the fill or stroke color. */
 int
 pdf_set_color(gx_device_pdf *pdev, gx_color_index color,
-  gx_color_index *pdcolor, const char *rgs)
-{	if ( *pdcolor != color )
-	  { FILE *file = pdev->file;
-	    float r = (color >> 16) / 255.0;
-	    float g = ((color >> 8) & 0xff) / 255.0;
-	    float b = (color & 0xff) / 255.0;
-	    if ( r == g && g == b )
-	      gprintf1(file, "%g ", r), fprintf(file, "%s\n", rgs + 1);
-	    else
-	      gprintf3(file, "%g %g %g ", r, g, b), fprintf(file, "%s\n", rgs);
-	    *pdcolor = color;
+  gx_drawing_color *pdcolor, const char *rgs)
+{	if ( gx_dc_pure_color(pdcolor) != color )
+	  { color_set_pure(pdcolor, color);
+	    psdf_set_color((gx_device_vector *)pdev, pdcolor, rgs);
 	  }
 	return 0;
 }
@@ -248,7 +233,8 @@ pdf_set_color(gx_device_pdf *pdev, gx_color_index color,
 /* Reset the graphics state parameters to initial values. */
 private void
 pdf_reset_graphics(gx_device_pdf *pdev)
-{	pdev->fill_color = pdev->stroke_color = 0;	/* black */
+{	color_set_pure(&pdev->fill_color, 0);	/* black */
+	color_set_pure(&pdev->stroke_color, 0);	/* ditto */
 	pdev->flatness = -1;
 	{ static const gx_line_params lp_initial = { gx_line_params_initial };
 	  pdev->line_params = lp_initial;
@@ -267,64 +253,72 @@ pdf_set_scale(gx_device_pdf *pdev)
 /* Begin a page contents part. */
 /* Return an error if the page has too many contents parts. */
 int
-pdf_begin_contents(gx_device_pdf *pdev)
-{	FILE *file = pdev->file;
-	int next = pdev->next_contents_id;
+pdf_open_contents(gx_device_pdf *pdev, pdf_context context)
+{	stream *s = pdev->strm;
 
-	switch ( pdev->in_contents )
-	{
-	case 2:			/* text */
-		fputs("ET\n", file);
-		pdev->in_contents = 1;
-	case 1:
-		return 0;
-	case 0:
-		;
-	}
+	if ( pdev->context == pdf_in_none )
+	  { int next = pdev->next_contents_id;
 
-	if ( next == max_contents_ids )
-	  return_error(gs_error_limitcheck);
-	pdev->contents_ids[next] = pdf_begin_obj(pdev);
-	pdev->next_contents_id = next + 1;
-	pdev->contents_length_id = pdf_obj_ref(pdev);
-	fprintf(file, "<< /Length %ld 0 R >>\n", pdev->contents_length_id);
-	fputs("stream\n", file);
-	pdev->contents_pos = ftell(file);
-	pdev->in_contents = 1;
-	if ( next == 0 )
-	  { /* Do a level of gsave for the clipping path. */
-	    fputs("q\n", file);
+	    if ( next == max_contents_ids )
+	      return_error(gs_error_limitcheck);
+	    pdev->contents_ids[next] = pdf_begin_obj(pdev);
+	    pdev->next_contents_id = next + 1;
+	    pdev->contents_length_id = pdf_obj_ref(pdev);
+	    pprintld1(s, "<< /Length %ld 0 R >>\n", pdev->contents_length_id);
+	    pputs(s, "stream\n");
+	    pdev->context = pdf_in_stream;
+	    pdev->contents_pos = stell(s);
+	    if ( next == 0 )
+	      { /* Do a level of gsave for the clipping path. */
+		pputs(s, "q\n");
+	      }
 	  }
+	if ( pdev->context == context )
+	  return 0;
+	switch ( context )
+	{
+	case pdf_in_text:
+	  /* pdev->context must be pdf_in_stream */
+	  pputs(s, "BT\n");
+	  pdev->procsets |= Text;
+	  break;
+	case pdf_in_stream:
+	  /* pdev->context must be pdf_in_text */
+	  pputs(s, "ET\n");
+	  break;
+	/*case pdf_in_none:*/	/* not possible */
+	}
+	pdev->context = context;
 	return 0;
 }
 
 /* Close the current contents part if we are in one. */
 int
 pdf_close_contents(gx_device_pdf *pdev, bool last)
-{	FILE *file = pdev->file;
+{	stream *s = pdev->strm;
 	long length;
 
-	switch ( pdev->in_contents )
+	switch ( pdev->context )
 	{
-	case 0:			/* not in contents */
+	case pdf_in_none:
 		return 0;
-	case 2:			/* in text */
-		fputs("ET\n", file);
-	case 1:			/* in graphics */
+	case pdf_in_text:
+		pputs(s, "ET\n");
+	case pdf_in_stream:
 		;
 	}
 	if ( last )
 	  { /* Exit from the clipping path gsave. */
-	    fputs("Q\n", file);
+	    pputs(s, "Q\n");
 	    pdev->text_state.font = 0;
 	  }
-	length = ftell(file) - pdev->contents_pos;
-	fputs("endstream\n", file);
+	length = stell(s) - pdev->contents_pos;
+	pputs(s, "endstream\n");
 	pdf_end_obj(pdev);
 	pdf_open_obj(pdev, pdev->contents_length_id);
-	fprintf(file, "%ld\n", length);
+	pprintld1(s, "%ld\n", length);
 	pdf_end_obj(pdev);
-	pdev->in_contents = 0;
+	pdev->context = pdf_in_none;
 	return 0;
 }
 
@@ -361,7 +355,9 @@ pdf_begin_aside(gx_device_pdf *pdev, pdf_resource **plist,
   const gs_memory_struct_type_t *pst, pdf_resource **ppres)
 {	pdf_resource *pres;
 
-	if ( pdev->in_contents && pdev->next_contents_id == max_contents_ids )
+	if ( pdev->context != pdf_in_none &&
+	     pdev->next_contents_id == max_contents_ids
+	   )
 	  return_error(gs_error_limitcheck);
 	if ( pst == NULL )
 	  pst = &st_pdf_resource;
@@ -383,14 +379,14 @@ pdf_begin_aside(gx_device_pdf *pdev, pdf_resource **plist,
 int
 pdf_begin_resource(gx_device_pdf *pdev, pdf_resource_type type,
   pdf_resource **ppres)
-{	FILE *file = pdev->file;
+{	stream *s = pdev->strm;
 	int code = pdf_begin_aside(pdev, &pdev->resources[type],
 				   resource_structs[type], ppres);
 
 	if ( code < 0 )
 	  return code;
-	fprintf(file, "<< /Type /%s /Name /R%ld",
-		resource_names[type], (*ppres)->id);
+	pprints1(s, "<< /Type /%s", resource_names[type]);
+	pprintld1(s, " /Name /R%ld", (*ppres)->id);
 	return code;
 }
 
@@ -412,7 +408,7 @@ void
 pdf_reset_page(gx_device_pdf *pdev)
 {	pdev->next_contents_id = 0;
 	pdf_reset_graphics(pdev);
-	pdev->procsets = 0;
+	pdev->procsets = NoMarks;
 	{ int i;
 	  for ( i = 0; i < num_resource_types; ++i )
 	    pdev->resources[i] = 0;
@@ -430,14 +426,28 @@ long
 pdf_page_id(gx_device_pdf *pdev, int page_num)
 {	long page_id;
 
-	if ( page_num >= 1 && page_num <= max_pages )
-	  { while ( page_num > pdev->pages_referenced )
-	      pdev->page_ids[pdev->pages_referenced++] = 0;
-	    if ( (page_id = pdev->page_ids[page_num - 1]) == 0 )
-	      pdev->page_ids[page_num - 1] = page_id = pdf_obj_ref(pdev);
+	if ( page_num >= pdev->num_page_ids )
+	  { /* Grow the page_ids array. */
+	    uint new_num_ids =
+	      max(page_num + 10, pdev->num_page_ids << 1);
+	    /* resize_object for a byte array takes a new object size */
+	    /* in bytes.  This is a quirk of the API that we probably */
+	    /* won't ever be able to fix.... */
+	    long *new_ids = gs_resize_object(pdev->pdf_memory, pdev->page_ids,
+					     new_num_ids * sizeof(long),
+					     "pdf_page_id(resize page_ids)");
+
+	    if ( new_ids == 0 )
+	      return 0;
+	    pdev->page_ids = new_ids;
+	    pdev->num_page_ids = new_num_ids;
 	  }
-	else
-	  page_id = 0;
+	if ( page_num < 1 )
+	  return 0;
+	while ( page_num > pdev->pages_referenced )
+	  pdev->page_ids[pdev->pages_referenced++] = 0;
+	if ( (page_id = pdev->page_ids[page_num - 1]) == 0 )
+	  pdev->page_ids[page_num - 1] = page_id = pdf_obj_ref(pdev);
 	return page_id;
 }
 
@@ -445,7 +455,7 @@ pdf_page_id(gx_device_pdf *pdev, int page_num)
 int
 pdf_write_saved_string(gx_device_pdf *pdev, gs_string *pstr)
 {	if ( pstr->data != 0 )
-	  { fwrite(pstr->data, 1, pstr->size, pdev->file);
+	  { pwrite(pdev->strm, pstr->data, pstr->size);
 	    gs_free_string(pdev->pdf_memory, pstr->data, pstr->size,
 			   "pdf_write_saved_string");
 	    pstr->data = 0;
@@ -455,61 +465,63 @@ pdf_write_saved_string(gx_device_pdf *pdev, gs_string *pstr)
 
 /* Open a page for writing. */
 int
-pdf_open_page(gx_device_pdf *pdev, bool contents)
-{	if ( !in_page(pdev) )
-	  { if ( pdev->next_page == max_pages )
-	      return_error(gs_error_limitcheck);
+pdf_open_page(gx_device_pdf *pdev, pdf_context context)
+{	if ( !is_in_page(pdev) )
+	  { if ( pdf_page_id(pdev, pdev->next_page + 1) == 0 )
+	      return_error(gs_error_VMerror);
 	    pdf_open_document(pdev);
 	  }
-	return (contents ? pdf_begin_contents(pdev) :
+	return (context != pdf_in_none ? pdf_open_contents(pdev, context) :
 		pdf_close_contents(pdev, false));
 }
 
 /* Close the current page. */
 private int
 pdf_close_page(gx_device_pdf *pdev)
-{	FILE *file = pdev->file;
+{	stream *s = pdev->strm;
 	int page_num = ++(pdev->next_page);
 	long page_id;
 
+	/* If the very first page is blank, we need to open the document */
+	/* before doing anything else. */
+	pdf_open_document(pdev);
 	pdf_close_contents(pdev, true);
 	page_id = pdf_page_id(pdev, page_num);
 	pdf_open_obj(pdev, page_id);
-	fprintf(file, "<<\n/Type /Page\n/MediaBox [%d %d %d %d]\n",
-		0, 0,
-		(int)(pdev->MediaSize[0]), (int)(pdev->MediaSize[1]));
-	fprintf(file, "/Parent %ld 0 R\n", pdev->pages_id);
-	fputs("/Resources << /ProcSet [/PDF", file);
+	pprintd2(s, "<<\n/Type /Page\n/MediaBox [0 0 %d %d]\n",
+		 (int)(pdev->MediaSize[0]), (int)(pdev->MediaSize[1]));
+	pprintld1(s, "/Parent %ld 0 R\n", pdev->pages_id);
+	pputs(s, "/Resources << /ProcSet [/PDF");
 	if ( pdev->procsets & ImageB )
-	  fputs(" /ImageB", file);
+	  pputs(s, " /ImageB");
 	if ( pdev->procsets & ImageC )
-	  fputs(" /ImageC", file);
+	  pputs(s, " /ImageC");
 	if ( pdev->procsets & ImageI )
-	  fputs(" /ImageI", file);
+	  pputs(s, " /ImageI");
 	if ( pdev->procsets & Text )
-	  fputs(" /Text", file);
-	fputs("]\n", file);
+	  pputs(s, " /Text");
+	pputs(s, "]\n");
 	{ int i;
 	  for ( i = 0; i < num_resource_types; ++i )
 	    { const pdf_resource *pres = pdev->resources[i];
 	      if ( pres != 0 )
-		{ fprintf(file, "/%s <<\n", resource_names[i]);
+		{ pprints1(s, "/%s <<\n", resource_names[i]);
 		  for ( ; pres; pres = pres->next )
-		    fprintf(file, "/R%ld %ld 0 R\n", pres->id, pres->id);
-		  fputs(">>\n", file);
+		    pprintld2(s, "/R%ld %ld 0 R\n", pres->id, pres->id);
+		  pputs(s, ">>\n");
 		}
 	      pdev->resources[i] = 0;
 	    }
 	}
-	fputs(">>\n", file);
+	pputs(s, ">>\n");
 	if ( pdev->next_contents_id == 1 )
-	  fprintf(file, "/Contents %ld 0 R\n", pdev->contents_ids[0]);
+	  pprintld1(s, "/Contents %ld 0 R\n", pdev->contents_ids[0]);
 	else
 	  { int i;
-	    fputs("/Contents [\n", file);
+	    pputs(s, "/Contents [\n");
 	    for ( i = 0; i < pdev->next_contents_id; ++i )
-	      fprintf(file, "%ld 0 R\n", pdev->contents_ids[i]);
-	    fputs("]\n", file);
+	      pprintld1(s, "%ld 0 R\n", pdev->contents_ids[i]);
+	    pputs(s, "]\n");
 	  }
 	pdf_write_saved_string(pdev, &pdev->page_string);
 	{ const pdf_resource *pres = pdev->annots;
@@ -517,15 +529,15 @@ pdf_close_page(gx_device_pdf *pdev)
 	  for ( ; pres != 0; pres = pres->next )
 	    if ( pres->rid == page_num - 1 )
 	      { if ( !any )
-		  { fputs("/Annots [\n", file);
+		  { pputs(s, "/Annots [\n");
 		    any = true;
 		  }
-		fprintf(file, "%ld 0 R\n", pres->id);
+		pprintld1(s, "%ld 0 R\n", pres->id);
 	      }
 	  if ( any )
-	    fputs("]\n", file);
+	    pputs(s, "]\n");
 	}
-	fputs(">>\n", file);
+	pputs(s, ">>\n");
 	pdf_end_obj(pdev);
 	pdf_reset_page(pdev);
 	return 0;
@@ -534,61 +546,23 @@ pdf_close_page(gx_device_pdf *pdev)
 /* Write the default entries of the Info dictionary. */
 int
 pdf_write_default_info(gx_device_pdf *pdev)
-{	FILE *file = pdev->file;
+{	stream *s = pdev->strm;
 	/* Reading the time without using time_t is a challenge.... */
 	long t[2];	/* time_t can't be longer than 2 longs. */
 	struct tm ltime;
+	char buf[20];
 
 	time((void *)t);
 	ltime = *localtime((void *)t);
-	fprintf(file, "/CreationDate (D:%04d%02d%02d%02d%02d%02d)\n",
+	sprintf(buf, "%04d%02d%02d%02d%02d%02d",
 		ltime.tm_year + 1900, ltime.tm_mon + 1, ltime.tm_mday,
 		ltime.tm_hour, ltime.tm_min, ltime.tm_sec);
-	fprintf(file, "/Producer (%s %1.2f)\n",
-		gs_product, gs_revision / 100.0);
+	pprints1(s, "/CreationDate (D:%s)\n", buf);
+	sprintf(buf, "%1.2f", gs_revision / 100.0);
+	pprints2(s, "/Producer (%s %s)\n", gs_product, buf);
 	return 0;
 }
 
-/* ------ Output ------ */
-
-/* Print (a) floating point number(s). */
-/* See gdevpdfx.h for why this is needed. */
-const char *
-gprintf1(FILE *file, const char *format, floatp v)
-{	const char *fp = format;
-	bool found = false;
-	char str[50];
-
-	sprintf(str, "%g", v);
-	if ( strchr(str, 'e') )
-	  { /* Bad news.  Try again using f-format. */
-	    sprintf(str, (fabs(v) > 1 ? "%1.1f" : "%1.8f"), v);
-	  }
-	for ( ; *fp != 0 && (*fp != '%' || !found); ++fp )
-	  { if ( *fp != '%' )
-	      { putc(*fp, file);
-	        continue;
-	      }
-#ifdef DEBUG
-	    if ( fp[1] != 'g' )		/* shouldn't happen! */
-	      lprintf1("Bad format in gprintf: %s\n", format);
-#endif
-	    fputs(str, file);
-	    found = true;
-	    ++fp;
-	  }
-	return fp;
-}
-const char *
-gprintf2(FILE *file, const char *format, floatp v1, floatp v2)
-{	return gprintf1(file, gprintf1(file, format, v1), v2);
-}
-const char *
-gprintf4(FILE *file, const char *format, floatp v1, floatp v2, floatp v3,
-  floatp v4)
-{	return gprintf2(file, gprintf2(file, format, v1, v2), v3, v4);
-}
-
 /* ---------------- Device open/close ---------------- */
 
 /* Open the device. */
@@ -596,6 +570,7 @@ private int
 pdf_open(gx_device *dev)
 {	gx_device_pdf *pdev = (gx_device_pdf *)dev;
 	char fmode[4];
+	int code;
 
 	pdev->pdf_memory = &gs_memory_default;		/* as good as any */
 	strcpy(fmode, "w+");
@@ -605,20 +580,25 @@ pdf_open(gx_device *dev)
 			       pdev->tfname, fmode);
 	if ( pdev->tfile == 0 )
 	  return_error(gs_error_invalidfileaccess);
-	pdev->file = gp_fopen(pdev->fname, gp_fmode_wb);
-	if ( pdev->file == 0 )
-	  { fclose(pdev->tfile);
-	    unlink(pdev->tfname);
-	    pdev->tfile = 0;
-	    return_error(gs_error_invalidfileaccess);
-	  }
+	code = gdev_vector_open_file((gx_device_vector *)pdev, sbuf_size);
+	if ( code < 0 )
+	  goto fail;
+	gdev_vector_init((gx_device_vector *)pdev);
+	/* Set in_page so the vector routines won't try to call */
+	/* any vector implementation procedures. */
+	pdev->in_page = true;
 	pdf_set_scale(pdev);
-	pdev->next_id = 1;
-	pdev->root_id = pdf_obj_ref(pdev);
-	pdev->pages_id = pdf_obj_ref(pdev);
+	pdf_initialize_ids(pdev);
 	pdev->outlines_id = 0;
 	pdev->next_page = 0;
-	pdev->clip_path_id = gs_no_id;
+	pdev->page_ids = (void *)
+	  gs_alloc_byte_array(pdev->pdf_memory, initial_num_page_ids,
+			      sizeof(*pdev->page_ids), "pdf_open(page_ids)");
+	if ( pdev->page_ids == 0 )
+	  { code = gs_error_VMerror;
+	    goto fail;
+	  }
+	pdev->num_page_ids = initial_num_page_ids;
 	pdev->pages_referenced = 0;
 	pdev->catalog_string.data = 0;
 	pdev->pages_string.data = 0;
@@ -632,6 +612,11 @@ pdf_open(gx_device *dev)
 	pdf_reset_page(pdev);
 
 	return 0;
+fail:
+	fclose(pdev->tfile);
+	unlink(pdev->tfname);
+	pdev->tfile = 0;
+	return code;
 }
 
 /* Wrap up ("output") a page. */
@@ -645,7 +630,7 @@ pdf_output_page(gx_device *dev, int num_copies, int flush)
 private int
 pdf_close(gx_device *dev)
 {	gx_device_pdf *pdev = (gx_device_pdf *)dev;
-	FILE *file = pdev->file;
+	stream *s = pdev->strm;
 	FILE *tfile = pdev->tfile;
 	long xref;
 	long named_dests_id = 0;
@@ -657,22 +642,22 @@ pdf_close(gx_device *dev)
 
 	/* Create the root (Catalog). */
 	pdf_open_obj(pdev, pdev->pages_id);
-	fputs("<< /Type /Pages /Kids [\n", file);
+	pputs(s, "<< /Type /Pages /Kids [\n");
 	{ int i;
 	  for ( i = 0; i < pdev->next_page; ++i )
-	    fprintf(file, "%ld 0 R\n", pdev->page_ids[i]);
+	    pprintld1(s, "%ld 0 R\n", pdev->page_ids[i]);
 	}
-	fprintf(file, "] /Count %d\n", pdev->next_page);
+	pprintd1(s, "] /Count %d\n", pdev->next_page);
 	pdf_write_saved_string(pdev, &pdev->pages_string);
-	fputs(">>\n", file);
+	pputs(s, ">>\n");
 	pdf_end_obj(pdev);
 	if ( pdev->outlines_id != 0 )
 	  { pdfmark_close_outline(pdev);	/* depth must be zero! */
 	    pdf_open_obj(pdev, pdev->outlines_id);
-	    fprintf(file, "<< /Count %d /First %ld 0 R /Last %ld 0 R >>\n",
-		    pdev->outlines_open,
-		    pdev->outline_levels[0].first.id,
-		    pdev->outline_levels[0].last.id);
+	    pprintd1(s, "<< /Count %d", pdev->outlines_open);
+	    pprintld2(s, " /First %ld 0 R /Last %ld 0 R >>\n",
+		      pdev->outline_levels[0].first.id,
+		      pdev->outline_levels[0].last.id);
 	    pdf_end_obj(pdev);
 	  }
 	if ( pdev->articles != 0 )
@@ -695,69 +680,77 @@ pdf_close(gx_device *dev)
 	if ( pdev->named_dests != 0 )
 	  { pdf_named_dest *pnd;
 	    named_dests_id = pdf_begin_obj(pdev);
-	    fputs("<<\n", file);
+	    pputs(s, "<<\n");
 	    while ( (pnd = pdev->named_dests) != 0 )
 	      { pdev->named_dests = pnd->next;
-		fwrite(pnd->key.data, 1, pnd->key.size, file);
-		fprintf(file, " %s\n", pnd->dest);
+		pwrite(s, pnd->key.data, pnd->key.size);
+		pprints1(s, " %s\n", pnd->dest);
 		gs_free_string(pdev->pdf_memory, pnd->key.data, pnd->key.size,
 			       "pdf_close(named_dest key)");
 		gs_free_object(pdev->pdf_memory, pnd, "pdf_close(named_dest)");
 	      }
-	    fputs(">>\n", file);
+	    pputs(s, ">>\n");
 	    pdf_end_obj(pdev);
 	  }
 	pdf_open_obj(pdev, pdev->root_id);
-	fprintf(file, "<< /Type /Catalog /Pages %ld 0 R\n", pdev->pages_id);
+	pprintld1(s, "<< /Type /Catalog /Pages %ld 0 R\n", pdev->pages_id);
 	if ( pdev->outlines_id != 0 )
-	  fprintf(file, "/Outlines %ld 0 R\n", pdev->outlines_id);
+	  pprintld1(s, "/Outlines %ld 0 R\n", pdev->outlines_id);
 	if ( pdev->articles != 0 )
 	  { pdf_article *part;
-	    fputs("/Threads [ ", file);
+	    pputs(s, "/Threads [ ");
 	    while ( (part = pdev->articles) != 0 )
 	      { pdev->articles = part->next;
-		fprintf(file, "%ld 0 R\n", part->id);
+		pprintld1(s, "%ld 0 R\n", part->id);
 		gs_free_string(pdev->pdf_memory, part->title.data,
 			       part->title.size, "pdf_close(article title)");
 		gs_free_object(pdev->pdf_memory, part, "pdf_close(article)");
 	      }
-	    fputs("]\n", file);
+	    pputs(s, "]\n");
 	  }
 	if ( named_dests_id != 0 )
-	  fprintf(file, "/Dests %ld 0 R\n", named_dests_id);
+	  pprintld1(s, "/Dests %ld 0 R\n", named_dests_id);
 	pdf_write_saved_string(pdev, &pdev->catalog_string);
-	fputs(">>\n", file);
+	pputs(s, ">>\n");
 	pdf_end_obj(pdev);
 
 	/* Create the Info directory. */
 	/* This is supposedly optional, but some readers may require it. */
 	if ( pdev->info_id == 0 )
 	  { pdev->info_id = pdf_begin_obj(pdev);
-	    fputs("<< ", file);
+	    pputs(s, "<< ");
 	    pdf_write_default_info(pdev);
-	    fputs(">>\n", file);
+	    pputs(s, ">>\n");
 	    pdf_end_obj(pdev);
 	}
 
 	/* Write the cross-reference section. */
-	xref = ftell(file);
-	fprintf(file, "xref\n0 %ld\n", pdev->next_id);
-	fputs("0000000000 65535 f \n", file);
-	fseek(tfile, 0, SEEK_SET);
+	xref = stell(s);
+	if ( pdev->FirstObjectNumber == 1 )
+	  pprintld1(s, "xref\n0 %ld\n0000000000 65535 f \n",
+		    pdev->next_id);
+	else
+	  pprintld2(s, "xref\n0 1\n0000000000 65535 f \n%ld %ld\n",
+		    pdev->FirstObjectNumber,
+		    pdev->next_id - pdev->FirstObjectNumber);
+	fseek(tfile, 0L, SEEK_SET);
 	{ long i;
-	  for ( i = 1; i < pdev->next_id; ++i )
+	  for ( i = pdev->FirstObjectNumber; i < pdev->next_id; ++i )
 	    { ulong pos;
+	      char str[21];
+
 	      fread(&pos, sizeof(pos), 1, tfile);
-	      fprintf(file, "%010ld 00000 n \n", pos);
+	      sprintf(str, "%010ld 00000 n \n", pos);
+	      pputs(s, str);
 	    }
 	}
 
 	/* Write the trailer. */
-	fputs("trailer\n", file);
-	fprintf(file, "<< /Size %ld /Root %ld 0 R /Info %ld 0 R\n",
-		pdev->next_id, pdev->root_id, pdev->info_id);
-	fputs(">>\n", file);
-	fprintf(file, "startxref\n%ld\n%%%%EOF\n", xref);
+	pputs(s, "trailer\n");
+	pprintld3(s, "<< /Size %ld /Root %ld 0 R /Info %ld 0 R\n",
+		  pdev->next_id, pdev->root_id, pdev->info_id);
+	pputs(s, ">>\n");
+	pprintld1(s, "startxref\n%ld\n%%%%EOF\n", xref);
 
 	/* Release the resource records. */
 	{ pdf_resource *pres;
@@ -769,8 +762,11 @@ pdf_close(gx_device *dev)
 	  pdev->last_resource = 0;
 	}
 
-	fclose(pdev->file);
-	pdev->file = 0;
+	gs_free_object(pdev->pdf_memory, pdev->page_ids, "page_ids");
+	pdev->page_ids = 0;
+	pdev->num_page_ids = 0;
+
+	gdev_vector_close_file((gx_device_vector *)pdev);
 	fclose(pdev->tfile);
 	pdev->tfile = 0;
 	unlink(pdev->tfname);
@@ -784,18 +780,20 @@ private int
 pdf_fill_rectangle(gx_device *dev, int x, int y, int w, int h,
   gx_color_index color)
 {	gx_device_pdf *pdev = (gx_device_pdf *)dev;
-	FILE *file = pdev->file;
+	stream *s = pdev->strm;
 	int code;
 
 	/* Make a special check for the initial fill with white, */
 	/* which shouldn't cause the page to be opened. */
-	if ( color == 0xffffff && !in_page(pdev) )
+	if ( color == 0xffffff && !is_in_page(pdev) )
 	  return 0;
-	code = pdf_open_page(pdev, true);
+	code = pdf_open_page(pdev, pdf_in_stream);
 	if ( code < 0 )
 	  return code;
+	/* Make sure we aren't being clipped. */
+	pdf_put_clip_path(pdev, NULL);
 	pdf_set_color(pdev, color, &pdev->fill_color, "rg");
-	gprintf4(file, "%g %g %g %g re\nf\n",
+	pprintg4(s, "%g %g %g %g re\nf\n",
 		 x / pdev->scale.x, y / pdev->scale.y,
 		 w / pdev->scale.x, h / pdev->scale.y);
 	return 0;
@@ -810,7 +808,7 @@ pdf_fill_rectangle(gx_device *dev, int x, int y, int w, int h,
 private int
 pdf_put_path(gx_device_pdf *pdev, const gx_path *ppath, bool do_close,
   const gs_matrix *pmat)
-{	FILE *file = pdev->file;
+{	stream *s = pdev->strm;
 	gs_fixed_rect rbox;
 	const subpath *next;
 	gs_path_enum cenum;
@@ -827,7 +825,7 @@ pdf_put_path(gx_device_pdf *pdev, const gx_path *ppath, bool do_close,
 		  { gs_point_transform_inverse(p.x, p.y, pmat, &p);
 		    gs_point_transform_inverse(q.x, q.y, pmat, &q);
 		  }
-		gprintf4(file, "%g %g %g %g re\n",
+		pprintg4(s, "%g %g %g %g re\n",
 			 p.x / pdev->scale.x, p.y / pdev->scale.y,
 			 (q.x - p.x) / pdev->scale.x,
 			 (q.y - p.y) / pdev->scale.y);
@@ -852,7 +850,7 @@ sw:	    switch ( pe_op )
 do1:		vp[0].x = fixed2float(vs[0]), vp[0].y = fixed2float(vs[1]);
 		if ( pmat )
 		  gs_point_transform_inverse(vp[0].x, vp[0].y, pmat, &vp[0]);
-		gprintf2(file, format,
+		pprintg2(s, format,
 			 vp[0].x / pdev->scale.x, vp[0].y / pdev->scale.y);
 		break;
 	      case gs_pe_curveto:
@@ -864,19 +862,19 @@ do1:		vp[0].x = fixed2float(vs[0]), vp[0].y = fixed2float(vs[1]);
 		    gs_point_transform_inverse(vp[1].x, vp[1].y, pmat, &vp[1]);
 		    gs_point_transform_inverse(vp[2].x, vp[2].y, pmat, &vp[2]);
 		  }
-		gprintf6(file, "%g %g %g %g %g %g c\n",
+		pprintg6(s, "%g %g %g %g %g %g c\n",
 			 vp[0].x / pdev->scale.x, vp[0].y / pdev->scale.y,
 			 vp[1].x / pdev->scale.x, vp[1].y / pdev->scale.y,
 			 vp[2].x / pdev->scale.x, vp[2].y / pdev->scale.y);
 		break;
 	      case gs_pe_closepath:
 		if ( do_close )
-		  { fputs("h\n", file);
+		  { pputs(s, "h\n");
 		    break;
 		  }
 		pe_op = gx_path_enum_next(&cenum, (gs_fixed_point *)vs);
 		if ( pe_op != 0 )
-		  { fputs("h\n", file);
+		  { pputs(s, "h\n");
 		    goto sw;
 		  }
 		return 1;
@@ -884,14 +882,44 @@ do1:		vp[0].x = fixed2float(vs[0]), vp[0].y = fixed2float(vs[1]);
 	  }
 }
 
+/* Test whether we will need to put the clipping path. */
+private bool
+pdf_must_put_clip_path(gx_device_pdf *pdev, const gx_clip_path *pcpath)
+{	if ( pcpath == NULL )
+	  return pdev->clip_path_id != pdev->no_clip_path_id;
+	if ( pdev->clip_path_id == pcpath->id )
+	  return false;
+	if ( gx_cpath_includes_rectangle(pcpath, fixed_0, fixed_0,
+					 int2fixed(pdev->width),
+					 int2fixed(pdev->height))
+	   )
+	  return pdev->clip_path_id != pdev->no_clip_path_id;
+	return true;
+}
+
 /* Put a clipping path on the output file. */
 int
 pdf_put_clip_path(gx_device_pdf *pdev, const gx_clip_path *pcpath)
-{	if ( pcpath != NULL && pcpath->id != pdev->clip_path_id )
-	  { FILE *file = pdev->file;
+{	stream *s = pdev->strm;
 
-	    fputs("Q\nq\nW\n", file);
-	    pdev->text_state.font = 0;
+	if ( pcpath == NULL ) {
+	  if ( pdev->clip_path_id == pdev->no_clip_path_id )
+	    return 0;
+	  pputs(s, "Q\nq\n");
+	  pdev->clip_path_id = pdev->no_clip_path_id;
+	} else {
+	  if ( pdev->clip_path_id == pcpath->id )
+	    return 0;
+	  if ( gx_cpath_includes_rectangle(pcpath, fixed_0, fixed_0,
+					   int2fixed(pdev->width),
+					   int2fixed(pdev->height))
+	     ) {
+	    if ( pdev->clip_path_id == pdev->no_clip_path_id )
+	      return 0;
+	    pputs(s, "Q\nq\n");
+	    pdev->clip_path_id = pdev->no_clip_path_id;
+	  } else {
+	    pputs(s, "Q\nq\nW\n");
 	    if ( pcpath->segments_valid )
 	      pdf_put_path(pdev, &pcpath->path, true, NULL);
 	    else
@@ -902,16 +930,20 @@ pdf_put_clip_path(gx_device_pdf *pdev, const gx_clip_path *pcpath)
 		  prect = &pcpath->list.single;
 		for ( ; prect != 0; prect = prect->next )
 		  if ( prect->xmax > prect->xmin && prect->ymax > prect->ymin )
-		    gprintf4(file, "%g %g %g %g re\n",
+		    pprintg4(s, "%g %g %g %g re\n",
 			     prect->xmin / pdev->scale.x,
 			     prect->ymin / pdev->scale.y,
 			     (prect->xmax - prect->xmin) / pdev->scale.x,
 			     (prect->ymax - prect->ymin) / pdev->scale.y);
 	      }
-	    fputs("n\n", file);
-	    pdf_reset_graphics(pdev);
+	    pputs(s, "n\n");
 	    pdev->clip_path_id = pcpath->id;
 	  }
+	}
+	pdev->text_state.font = 0;
+	if ( pdev->context == pdf_in_text )
+	  pdev->context = pdf_in_stream;
+	pdf_reset_graphics(pdev);
 	return 0;
 }
 
@@ -923,33 +955,48 @@ pdf_fill_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
   const gx_fill_params *params,
   const gx_drawing_color *pdcolor, const gx_clip_path *pcpath)
 {	gx_device_pdf *pdev = (gx_device_pdf *)dev;
-	FILE *file = pdev->file;
+	stream *s = pdev->strm;
 	int code;
+	/*
+	 * HACK: we fill an empty path in order to set the clipping path
+	 * and the color for writing text.  If it weren't for this, we
+	 * could detect and skip empty paths before putting out the clip
+	 * path or the color.
+	 */
+	bool have_path = !gx_path_is_void(ppath);
 
 	if ( !gx_dc_is_pure(pdcolor) )
 	  return gx_default_fill_path(dev, pis, ppath, params, pdcolor,
 				      pcpath);
-	/* Make a special check for the initial fill with white, */
-	/* which shouldn't cause the page to be opened. */
-	if ( gx_dc_pure_color(pdcolor) == 0xffffff && !in_page(pdev) )
+	/*
+	 * Make a special check for the initial fill with white,
+	 * which shouldn't cause the page to be opened.
+	 */
+	if ( gx_dc_pure_color(pdcolor) == 0xffffff && !is_in_page(pdev) )
 	  return 0;
-	code = pdf_open_page(pdev, true);
-	if ( code < 0 )
-	  return code;
+	if ( have_path || pdev->context == pdf_in_none ||
+	     pdf_must_put_clip_path(pdev, pcpath)
+	   )
+	  { code = pdf_open_page(pdev, pdf_in_stream);
+	    if ( code < 0 )
+	      return code;
+	  }
 	pdf_put_clip_path(pdev, pcpath);
 	pdf_set_color(pdev, gx_dc_pure_color(pdcolor), &pdev->fill_color, "rg");
-	if ( params->flatness != pdev->flatness )
-	  { gprintf1(file, "%g i\n", params->flatness);
-	    pdev->flatness = params->flatness;
+	if ( have_path )
+	  { if ( params->flatness != pdev->flatness )
+	      { pprintg1(s, "%g i\n", params->flatness);
+	        pdev->flatness = params->flatness;
+	      }
+	    pdf_put_path(pdev, ppath, false, NULL);
+	    pputs(s, (params->rule < 0 ? "f\n" : "f*\n"));
 	  }
-	pdf_put_path(pdev, ppath, false, NULL);
-	fprintf(file, "%s\n", (params->rule < 0 ? "f" : "f*"));
 	return 0;
 }
 
 /* Compare two dash patterns. */
 private bool
-dash_pattern_eq(const float *stored, const gx_dash_params *set, float scale)
+pdf_dash_pattern_eq(const float *stored, const gx_dash_params *set, float scale)
 {	int i;
 	for ( i = 0; i < set->pattern_size; ++i )
 	  if ( stored[i] != (float)(set->pattern[i] * scale) )
@@ -963,7 +1010,7 @@ pdf_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
   const gx_stroke_params *params,
   const gx_drawing_color *pdcolor, const gx_clip_path *pcpath)
 {	gx_device_pdf *pdev = (gx_device_pdf *)dev;
-	FILE *file = pdev->file;
+	stream *s = pdev->strm;
 	int code;
 	int pattern_size = pis->line_params.dash.pattern_size;
 	double scale;
@@ -972,7 +1019,7 @@ pdf_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 	if ( !gx_dc_is_pure(pdcolor) )
 	  return gx_default_stroke_path(dev, pis, ppath, params, pdcolor,
 					pcpath);
-	code = pdf_open_page(pdev, true);
+	code = pdf_open_page(pdev, pdf_in_stream);
 	if ( code < 0 )
 	  return code;
 	/* Check for a CTM we can handle. */
@@ -1005,45 +1052,45 @@ pdf_stroke_path(gx_device *dev, const gs_imager_state *pis, gx_path *ppath,
 	     pattern_size != pdev->line_params.dash.pattern_size ||
 	     pattern_size > max_dash ||
 	     (pattern_size != 0 &&
-	      !dash_pattern_eq(pdev->dash_pattern, &pis->line_params.dash,
-			       scale))
+	      !pdf_dash_pattern_eq(pdev->dash_pattern, &pis->line_params.dash,
+				   scale))
 	   )
-	  { fputs("[ ", file);
+	  { pputs(s, "[ ");
 	    pdev->line_params.dash.pattern_size = pattern_size;
 	    for ( i = 0; i < pattern_size; ++i )
 	      { float element = pis->line_params.dash.pattern[i] * scale;
 		if ( i < max_dash )
 		  pdev->dash_pattern[i] = element;
-		gprintf1(file, "%g ", element);
+		pprintg1(s, "%g ", element);
 	      }
 	    pdev->line_params.dash.offset =
 	      pis->line_params.dash.offset * scale;
-	    gprintf1(file, "] %g d\n", pdev->line_params.dash.offset);
+	    pprintg1(s, "] %g d\n", pdev->line_params.dash.offset);
 	  }
 	if ( params->flatness != pdev->flatness )
-	  { gprintf1(file, "%g i\n", params->flatness);
+	  { pprintg1(s, "%g i\n", params->flatness);
 	    pdev->flatness = params->flatness;
 	  }
 	if ( (float)(pis->line_params.half_width * scale) != pdev->line_params.half_width )
 	  { pdev->line_params.half_width = pis->line_params.half_width * scale;
-	    gprintf1(file, "%g w\n", pdev->line_params.half_width * 2);
+	    pprintg1(s, "%g w\n", pdev->line_params.half_width * 2);
 	  }
 	if ( pis->line_params.miter_limit != pdev->line_params.miter_limit )
-	  { gprintf1(file, "%g M\n", pis->line_params.miter_limit);
+	  { pprintg1(s, "%g M\n", pis->line_params.miter_limit);
 	    gx_set_miter_limit(&pdev->line_params,
 			       pis->line_params.miter_limit);
 	  }
 	if ( pis->line_params.cap != pdev->line_params.cap )
-	  { fprintf(file, "%d J\n", pis->line_params.cap);
+	  { pprintd1(s, "%d J\n", pis->line_params.cap);
 	    pdev->line_params.cap = pis->line_params.cap;
 	  }
 	if ( pis->line_params.join != pdev->line_params.join )
-	  { fprintf(file, "%d j\n", pis->line_params.join);
+	  { pprintd1(s, "%d j\n", pis->line_params.join);
 	    pdev->line_params.join = pis->line_params.join;
 	  }
 	code = pdf_put_path(pdev, ppath, false, NULL);
 	if ( code < 0 )
 	  return code;
-	fputs((code ? "s\n" : "S\n"), file);
+	pputs(s, (code ? "s\n" : "S\n"));
 	return 0;
 }
