@@ -1,24 +1,19 @@
-/* Copyright (C) 1990, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1990, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* gdevprn.c */
@@ -71,10 +66,9 @@ gdev_prn_alloc(gx_device *pdev)
 
 	memset(ppdev->skip, 0, sizeof(ppdev->skip));
 	ppdev->orig_procs = pdev->std_procs;
-	ppdev->ccfile = ppdev->cbfile = NULL;
 	mem_space = gdev_mem_bitmap_size(pmemdev);
 
-	if ( mem_space >= ppdev->max_bitmap ||
+	if ( mem_space >= ppdev->space_params.MaxBitmap ||
 	     mem_space != (uint)mem_space ||	/* too big to allocate */
 	     (base = (byte *)gs_malloc((uint)mem_space, 1, "printer buffer")) == 0 ||	/* can't allocate */
 	     (PRN_MIN_MEMORY_LEFT != 0 &&
@@ -82,11 +76,11 @@ gdev_prn_alloc(gx_device *pdev)
 	   )
 	{	/* Buffer the image in a command list. */
 		uint space;
-		int code, bcode = 0, ccode = 0;
+		int code;
 
 		/* Release the buffer if we allocated it. */
 		gs_free(base, (uint)mem_space, 1, "printer buffer(open)");
-		for ( space = ppdev->use_buffer_space; ; )
+		for ( space = ppdev->space_params.BufferSpace; ; )
 		{	base = (byte *)gs_malloc(space, 1,
 						 "command list buffer");
 			if ( base != 0 ) break;
@@ -98,17 +92,19 @@ open_c:		pcldev->data = base;
 		pcldev->target = pdev;
 		pcldev->make_buffer_device =
 		  ppdev->printer_procs.make_buffer_device;
+		pcldev->band_params = ppdev->space_params.band;
 		ppdev->buf = base;
 		ppdev->buffer_space = space;
 
 		/* Try opening the command list, to see if we allocated */
 		/* enough buffer space. */
-		code = (*gs_clist_device_procs.open_device)((gx_device *)pcldev);
+		code =
+		  (*gs_clist_device_procs.open_device)((gx_device *)pcldev);
 		if ( code < 0 )
 		{	/* If there wasn't enough room, and we haven't */
 			/* already shrunk the buffer, try enlarging it. */
 			if ( code == gs_error_limitcheck &&
-			     space >= ppdev->use_buffer_space
+			     space >= ppdev->space_params.BufferSpace
 			   )
 			{	gs_free(base, space, 1,
 					"command list buffer(retry open)");
@@ -119,21 +115,10 @@ open_c:		pcldev->data = base;
 				if ( base != 0 )
 				  goto open_c;
 			}
-			/* Fall through with code < 0 */
-		}
-		if ( code < 0 ||
-		     (ccode = clist_open_scratch(ppdev->ccfname, &ppdev->ccfile, &gs_memory_default, true)) < 0 ||
-		     (bcode = clist_open_scratch(ppdev->cbfname, &ppdev->cbfile, &gs_memory_default, false)) < 0
-		   )
-		{	/* Clean up before exiting */
-			eprintf3("Problem opening clist: code=%d, bcode=%d, ccode=%d\n",
-				 code, bcode, ccode);
+			/* Failure.  Clean up before exiting. */
 			gdev_prn_free(pdev);
-			return (code < 0 ? code : ccode < 0 ? ccode : bcode);
+			return code;
 		}
-		pcldev->cfile = ppdev->ccfile;
-		pcldev->bfile = ppdev->cbfile;
-		pcldev->bfile_end_pos = 0;
 		ppdev->std_procs = gs_clist_device_procs;
 	}
 	else
@@ -171,8 +156,11 @@ open_c:		pcldev->data = base;
 	/* standard macros for generating their procedure vectors. */
 	set_dev_proc(ppdev, get_page_device, gx_page_device_get_page_device);
 	copy_proc(get_alpha_bits);
+	copy_proc(get_clipping_box);
 #undef copy_proc
-	return (*dev_proc(pdev, open_device))(pdev);
+	/* If we're using a command list, we already opened the device. */
+	return (ppdev->buffer_space ?  0 :
+		(*dev_proc(pdev, open_device))(pdev));
 }
 
 /* Generic closing for the printer device. */
@@ -192,16 +180,10 @@ gdev_prn_close(gx_device *pdev)
 /* Again, we break this out for resetting page dimensions. */
 private int
 gdev_prn_free(gx_device *pdev)
-{	if ( ppdev->ccfile != NULL )
-	{	clist_fclose_and_unlink(ppdev->ccfile, ppdev->ccfname);
-		ppdev->ccfile = NULL;
-	}
-	if ( ppdev->cbfile != NULL )
-	{	clist_fclose_and_unlink(ppdev->cbfile, ppdev->cbfname);
-		ppdev->cbfile = NULL;
-	}
-	if ( ppdev->buffer_space != 0 )
-	{	/* Free the buffer */
+{	if ( ppdev->buffer_space != 0 )
+	{	/* Close files and release storage. */
+		(*gs_clist_device_procs.close_device)((gx_device *)pcldev);
+		/* Free the buffer */
 		gs_free(ppdev->buf, (uint)ppdev->buffer_space, 1,
 			"command list buffer(close)");
 		ppdev->buf = 0;
@@ -226,28 +208,22 @@ gdev_prn_get_params(gx_device *pdev, gs_param_list *plist)
 {	int code = gx_default_get_params(pdev, plist);
 	gs_param_string ofns;
 
-	if ( code < 0 ) return code;
-
-	code = (ppdev->NumCopies_set ?
-		param_write_int(plist, "NumCopies", &ppdev->NumCopies) :
-		param_write_null(plist, "NumCopies"));
-	if ( code < 0 ) return code;
-
-	code = param_write_bool(plist, "OpenOutputFile", &ppdev->OpenOutputFile);
-	if ( code < 0 ) return code;
-
-	if ( ppdev->Duplex_set >= 0 )
-	  { code =
-	      (ppdev->Duplex_set ? param_write_bool(plist, "Duplex", &ppdev->Duplex) :
-	       param_write_null(plist, "Duplex"));
-	    if ( code < 0 ) return code;
-	  }
-
-	code = param_write_long(plist, "BufferSpace", &ppdev->use_buffer_space);
-	if ( code < 0 ) return code;
-
-	code = param_write_long(plist, "MaxBitmap", &ppdev->max_bitmap);
-	if ( code < 0 ) return code;
+	if ( code < 0 ||
+	     (code = param_write_long(plist, "MaxBitmap", &ppdev->space_params.MaxBitmap)) < 0 ||
+	     (code = param_write_long(plist, "BufferSpace", &ppdev->space_params.BufferSpace)) < 0 ||
+	     (code = param_write_int(plist, "BandWidth", &ppdev->space_params.band.BandWidth)) < 0 ||
+	     (code = param_write_int(plist, "BandHeight", &ppdev->space_params.band.BandHeight)) < 0 ||
+	     (code = param_write_long(plist, "BandBufferSpace", &ppdev->space_params.band.BandBufferSpace)) < 0 ||
+	     (code = (ppdev->NumCopies_set ?
+		      param_write_int(plist, "NumCopies", &ppdev->NumCopies) :
+		      param_write_null(plist, "NumCopies"))) < 0 ||
+	     (code = param_write_bool(plist, "OpenOutputFile", &ppdev->OpenOutputFile)) < 0 ||
+	     (ppdev->Duplex_set >= 0 &&
+	      (code = (ppdev->Duplex_set ?
+		       param_write_bool(plist, "Duplex", &ppdev->Duplex) :
+		       param_write_null(plist, "Duplex"))) < 0)
+	   )
+	  return code;
 
 	ofns.data = (const byte *)ppdev->fname,
 	  ofns.size = strlen(ppdev->fname),
@@ -266,16 +242,15 @@ gdev_prn_put_params(gx_device *pdev, gs_param_list *plist)
 	bool ncset = ppdev->NumCopies_set;
 	bool oof = ppdev->OpenOutputFile;
 	bool duplex;
-	  int duplex_set = -1;
+	int duplex_set = -1;
 	int width = pdev->width;
 	int height = pdev->height;
-	long buffer_space = ppdev->use_buffer_space;
-	long max_bitmap = ppdev->max_bitmap;
-	long bsl = buffer_space;
-	long mbl = max_bitmap;
+	gdev_prn_space_params sp, save_sp;
 	gs_param_string ofs;
 	gs_param_dict mdict;
 
+	sp = ppdev->space_params;
+	save_sp = sp;
 	switch ( code = param_read_int(plist, (param_name = "NumCopies"), &nci) )
 	{
 	case 0:
@@ -325,34 +300,35 @@ nce:		param_signal_error(plist, param_name, ecode);
 		;
 	    }
 
-	switch ( code = param_read_long(plist, (param_name = "BufferSpace"), &bsl) )
+#define CHECK_PARAM_CASES(bad, label)\
+    case 0: if ( bad ) ecode = gs_error_rangecheck; else break; goto label;\
+    default: ecode = code;\
+  label: param_signal_error(plist, param_name, ecode);\
+    case 1: break
+
+	switch ( code = param_read_long(plist, (param_name = "MaxBitmap"), &sp.MaxBitmap) )
 	{
-	case 0:
-		if ( bsl < 10000 )
-		  ecode = gs_error_rangecheck;
-		else
-		  break;
-		goto bse;
-	default:
-		ecode = code;
-bse:		param_signal_error(plist, param_name, ecode);
-	case 1:
-		break;
+	  CHECK_PARAM_CASES(sp.MaxBitmap < 10000, mbe);
 	}
 
-	switch ( code = param_read_long(plist, (param_name = "MaxBitmap"), &mbl) )
+	switch ( code = param_read_long(plist, (param_name = "BufferSpace"), &sp.BufferSpace) )
 	{
-	case 0:
-		if ( mbl < 10000 )
-		  ecode = gs_error_rangecheck;
-		else
-		  break;
-		goto mbe;
-	default:
-		ecode = code;
-mbe:		param_signal_error(plist, param_name, ecode);
-	case 1:
-		break;
+	  CHECK_PARAM_CASES(sp.BufferSpace < 10000, bse);
+	}
+
+	switch ( code = param_read_int(plist, (param_name = "BandWidth"), &sp.band.BandWidth) )
+	{
+	  CHECK_PARAM_CASES(sp.band.BandWidth < 0, bwe);
+	}
+
+	switch ( code = param_read_int(plist, (param_name = "BandHeight"), &sp.band.BandHeight) )
+	{
+	  CHECK_PARAM_CASES(sp.band.BandHeight < 0, bhe);
+	}
+
+	switch ( code = param_read_long(plist, (param_name = "BandBufferSpace"), &sp.band.BandBufferSpace) )
+	{
+	  CHECK_PARAM_CASES(sp.band.BandBufferSpace < 0, bbse);
 	}
 
 	switch ( code = param_read_string(plist, (param_name = "OutputFile"), &ofs) )
@@ -436,19 +412,17 @@ ofe:		param_signal_error(plist, param_name, ecode);
 	    ppdev->Duplex_set = duplex_set;
 	  }
 	/* If necessary, free and reallocate the printer memory. */
-	if ( bsl != buffer_space || mbl != max_bitmap ||
+	if ( memcmp(&sp, &save_sp, sizeof(sp)) != 0 ||
 	     pdev->width != width || pdev->height != height
 	   )
 	  {	if ( is_open )
 		  gdev_prn_free(pdev);
-		ppdev->use_buffer_space = bsl;
-		ppdev->max_bitmap = mbl;
+		ppdev->space_params = sp;
 		if ( is_open )
 		  {	ecode = code = gdev_prn_alloc(pdev);
 			if ( code < 0 )
 			  {	/* Try to put things back as they were. */
-				ppdev->use_buffer_space = buffer_space;
-				ppdev->max_bitmap = max_bitmap;
+				ppdev->space_params = save_sp;
 				gx_device_set_width_height(pdev,
 							   width, height);
 				code = gdev_prn_alloc(pdev);
@@ -570,34 +544,35 @@ gdev_prn_output_media(int index, gs_param_dict *pdict, const output_media *pom)
 /* Generic routine to send the page to the printer. */
 int
 gdev_prn_output_page(gx_device *pdev, int num_copies, int flush)
-{	int code, outcode, closecode, errcode;
+{	int outcode, closecode, errcode, endcode;
 
 	if ( num_copies > 0 )
-	  {	code = gdev_prn_open_printer(pdev, 1);
-		if ( code < 0 )
-		  return code;
+	  { int code = gdev_prn_open_printer(pdev, 1);
+	    if ( code < 0 )
+	      return code;
 
-		/* Print the accumulated page description. */
-		outcode =
-		  (*ppdev->printer_procs.print_page_copies)(ppdev,
-							    ppdev->file,
-							    num_copies);
-		errcode = (ferror(ppdev->file) ? gs_error_ioerror : 0);
-		closecode = gdev_prn_close_printer(pdev);
+	    /* Print the accumulated page description. */
+	    outcode =
+	      (*ppdev->printer_procs.print_page_copies)(ppdev, ppdev->file,
+							num_copies);
+	    fflush(ppdev->file);
+	    errcode =
+	      (ferror(ppdev->file) ? gs_note_error(gs_error_ioerror) : 0);
+	    closecode = gdev_prn_close_printer(pdev);
 	  }
 	else
-	  code = outcode = closecode = errcode = 0;
-
-	if ( ppdev->buffer_space ) /* reinitialize clist for writing */
-	  code = (*gs_clist_device_procs.output_page)(pdev, num_copies, flush);
+	  outcode = closecode = errcode = 0;
+	endcode = (ppdev->buffer_space ? clist_finish_page(pdev, flush) : 0);
 
 	if ( outcode < 0 )
 	  return outcode;
 	if ( errcode < 0 )
-	  return_error(errcode);
+	  return errcode;
 	if ( closecode < 0 )
 	  return closecode;
-	return code;
+	if ( endcode < 0 )
+	  return endcode;
+	return outcode;
 }
 
 /* Print multiple copies of a page by calling print_page multiple times. */
