@@ -1,24 +1,19 @@
 /* Copyright (C) 1991, 1995, 1996 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* gdevsvga.c */
@@ -913,3 +908,45 @@ tseng_set_page(gx_device_svga *dev, int pn, int wnum)
 	if ( wnum ) pn <<= shift, mask <<= shift;
 	outportb(0x3cd, (inportb(0x3cd) & ~mask) + pn);
 }
+
+/* ------ The Avance Logic device (mostly experimental) ------ */
+/* For questions about this device, please contact Stefan Freund */
+/* <freund@ikp.uni-koeln.de>. */
+
+private dev_proc_open_device(ali_open);
+private gx_device_procs ali_procs = svga_procs(ali_open);
+/* We can use the atiw_get/set_mode procedures. */
+private void ali_set_page(P3(gx_device_svga *, int, int));
+
+/* The 256-color Avance Logic device */
+gx_device_svga gs_ali_device =
+	svga_device(ali_procs, "ali", atiw_get_mode, atiw_set_mode,
+ali_set_page);
+
+/* Initialize the graphics mode. */
+private int
+ali_open(gx_device *dev)
+{	fb_dev->wnum_read = 1;
+	fb_dev->wnum_write = 0;
+	/* Select the proper video mode */
+	   {	static const mode_info mode_table[] = {
+		   {	 640,  400, 0x29	},
+		   {	 640,  480, 0x2a	},
+		   {	 800,  600, 0x2c	},
+		   {	 1024, 768, 0x31	},
+		   {	-1, -1, -1	}
+		};
+		int code = svga_find_mode(dev, mode_table);
+		if ( code < 0 ) return code;	/* mode not available */
+		return svga_open(dev);
+	   }
+
+}
+
+/* Set the current display page. */
+private void
+ali_set_page(gx_device_svga *dev, int pn, int wnum)
+{
+	outportb(0x3d6, pn);   /* read	*/
+	outportb(0x3d7, pn);   /* write */
+}
