@@ -1,24 +1,19 @@
-/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* gschar.c */
@@ -57,21 +52,22 @@
 private_st_gs_show_enum();
 #define eptr ((gs_show_enum *)vptr)
 private ENUM_PTRS_BEGIN(show_enum_enum_ptrs) {
-	if ( index > eptr->fstack.depth + 5 )
+	if ( index > eptr->fstack.depth + 6 )
 	  return 0;
-	*pep = eptr->fstack.items[index - 5].font;
-	break;
+	ENUM_RETURN(eptr->fstack.items[index - 6].font);
 	}
 	ENUM_PTR(0, gs_show_enum, pgs);
 	ENUM_CONST_STRING_PTR(1, gs_show_enum, str);
-	ENUM_PTR(2, gs_show_enum, dev_cache);
-	ENUM_PTR(3, gs_show_enum, dev_cache2);
-	ENUM_PTR(4, gs_show_enum, dev_null);
+	ENUM_PTR(2, gs_show_enum, show_gstate);
+	ENUM_PTR(3, gs_show_enum, dev_cache);
+	ENUM_PTR(4, gs_show_enum, dev_cache2);
+	ENUM_PTR(5, gs_show_enum, dev_null);
 ENUM_PTRS_END
 private RELOC_PTRS_BEGIN(show_enum_reloc_ptrs) {
 	int i;
 	RELOC_PTR(gs_show_enum, pgs);
 	RELOC_CONST_STRING_PTR(gs_show_enum, str);
+	RELOC_PTR(gs_show_enum, show_gstate);
 	RELOC_PTR(gs_show_enum, dev_cache);
 	RELOC_PTR(gs_show_enum, dev_cache2);
 	RELOC_PTR(gs_show_enum, dev_null);
@@ -84,7 +80,7 @@ private RELOC_PTRS_BEGIN(show_enum_reloc_ptrs) {
 private int continue_kshow(P1(gs_show_enum *));
 private int continue_show(P1(gs_show_enum *));
 private int continue_show_update(P1(gs_show_enum *));
-private int show_setup(P3(gs_show_enum *, gs_state *, const char *));
+private int show_setup(P4(gs_show_enum *, gs_state *, const char *, bool));
 private void show_set_scale(P1(gs_show_enum *));
 private int show_cache_setup(P1(gs_show_enum *));
 private int show_state_setup(P1(gs_show_enum *));
@@ -175,7 +171,7 @@ gs_show_n_init(register gs_show_enum *penum,
    gs_state *pgs, const char *str, uint size)
 {	setup_show_n();
 	penum->slow_show = false;
-	return show_setup(penum, pgs, str);
+	return show_setup(penum, pgs, str, true);
 }
 int
 gs_show_init(gs_show_enum *penum, gs_state *pgs, const char *str)
@@ -188,7 +184,7 @@ gs_ashow_n_init(register gs_show_enum *penum,
   gs_state *pgs, floatp ax, floatp ay, const char *str, uint size)
 {	int code;
 	setup_show_n();
-	code = show_setup(penum, pgs, str);
+	code = show_setup(penum, pgs, str, true);
 	setup_a();
 	return code;
 }
@@ -204,7 +200,7 @@ gs_widthshow_n_init(register gs_show_enum *penum,
   gs_state *pgs, floatp cx, floatp cy, gs_char chr, const char *str, uint size)
 {	int code;
 	setup_show_n();
-	code = show_setup(penum, pgs, str);
+	code = show_setup(penum, pgs, str, true);
 	setup_width();
 	return code;
 }
@@ -221,7 +217,7 @@ gs_awidthshow_n_init(register gs_show_enum *penum,
   const char *str, uint size)
 {	int code;
 	setup_show_n();
-	code = show_setup(penum, pgs, str);
+	code = show_setup(penum, pgs, str, true);
 	setup_a();
 	setup_width();
 	return code;
@@ -242,7 +238,7 @@ gs_kshow_n_init(register gs_show_enum *penum,
 	if ( pgs->font->FontType == ft_composite)
 	  return_error(gs_error_invalidfont);
 	setup_show_n();
-	code = show_setup(penum, pgs, str);
+	code = show_setup(penum, pgs, str, true);
 	penum->do_kern = 1;
 	penum->slow_show = true;
 	return code;
@@ -258,7 +254,7 @@ gs_xyshow_n_init(register gs_show_enum *penum,
    gs_state *pgs, const char *str, uint size)
 {	int code;
 	setup_show_n();
-	code = show_setup(penum, pgs, str);
+	code = show_setup(penum, pgs, str, true);
 	penum->do_kern = -1;
 	penum->slow_show = true;
 	return code;
@@ -269,15 +265,31 @@ gs_xyshow_init(gs_show_enum *penum, gs_state *pgs, const char *str)
 }
 
 /* glyphshow */
+private int
+setup_glyph(P3(gs_show_enum *, gs_state *, gs_glyph));
 private font_proc_encode_char(gs_glyphshow_encode_char);
 int
 gs_glyphshow_init(gs_show_enum *penum, gs_state *pgs, gs_glyph glyph)
+{	return setup_glyph(penum, pgs, glyph);
+}
+int
+gs_glyphpath_init(gs_show_enum *penum, gs_state *pgs, gs_glyph glyph,
+  bool stroke_path)
+{	int code = setup_glyph(penum, pgs, glyph);
+	penum->charpath_flag =
+	  (stroke_path ? cpm_true_charpath : cpm_false_charpath);
+	penum->can_cache = -1;
+	if_debug1('k', "[k]glyphpath, can_cache=%d", penum->can_cache);
+	return code;
+}
+private int
+setup_glyph(gs_show_enum *penum, gs_state *pgs, gs_glyph glyph)
 {	int code;
 	if ( pgs->font->FontType == ft_composite)
 	  return_error(gs_error_invalidfont);
 	penum->str.size = 1;
 	penum->slow_show = false;
-	code = show_setup(penum, pgs, "\000");	/* arbitrary char */
+	code = show_setup(penum, pgs, "\000", true); /* arbitrary char */
 	penum->current_glyph = glyph;
 	penum->encode_char = gs_glyphshow_encode_char;
 	return code;
@@ -297,7 +309,7 @@ gs_cshow_n_init(register gs_show_enum *penum,
    gs_state *pgs, const char *str, uint size)
 {	int code;
 	setup_show_n();
-	code = show_setup(penum, pgs, str);
+	code = show_setup(penum, pgs, str, false);
 	penum->do_kern = -1;
 	penum->stringwidth_flag = -1;
 	penum->slow_show = true;
@@ -322,7 +334,8 @@ gs_stringwidth_init(gs_show_enum *penum, gs_state *pgs, const char *str)
 /* Common code for stringwidth[_n] */
 private int
 stringwidth_setup(gs_show_enum *penum, gs_state *pgs, const char *str)
-{	int code = (penum->slow_show = false, show_setup(penum, pgs, str));
+{	int code =
+	  (penum->slow_show = false, show_setup(penum, pgs, str, false));
 	gs_memory_t *mem = pgs->memory;
 	gx_device_null *dev_null;
 	if ( code < 0 )
@@ -354,7 +367,7 @@ gs_charpath_n_init(gs_show_enum *penum, gs_state *pgs,
   const char *str, uint size, bool stroke_path)
 {	int code;
 	setup_show_n();
-	code = show_setup(penum, pgs, str);
+	code = show_setup(penum, pgs, str, false);
 	penum->charpath_flag =
 	  (stroke_path ? cpm_true_charpath : cpm_false_charpath);
 	penum->can_cache = -1;
@@ -373,7 +386,7 @@ gs_charboxpath_n_init(gs_show_enum *penum, gs_state *pgs,
   const char *str, uint size, bool use_boxes)
 {	int code;
 	setup_show_n();
-	code = show_setup(penum, pgs, str);
+	code = show_setup(penum, pgs, str, false);
 	penum->charpath_flag =
 	  (use_boxes ? cpm_true_charboxpath : cpm_false_charboxpath);
 	penum->can_cache = 0;		/* different from charpath! */
@@ -652,8 +665,6 @@ gs_show_next(gs_show_enum *penum)
 }
 
 /* Continuation procedures */
-#define show_fast_move(wxy, pgs)\
-  gx_path_add_rel_point_inline(pgs->path, wxy.x, wxy.y)
 private int show_update(P1(gs_show_enum *penum));
 private int show_move(P1(gs_show_enum *penum));
 private int show_proceed(P1(gs_show_enum *penum));
@@ -725,7 +736,7 @@ show_update(register gs_show_enum *penum)
 		code = gs_grestore(pgs);
 		if ( code < 0 )
 		  return code;
-		code = gx_color_load(pgs->dev_color, pgs);
+		code = gs_state_color_load(pgs);
 		if ( code < 0 )
 		  return code;
 		return gx_image_cached_char(penum, cc);
@@ -745,6 +756,14 @@ show_update(register gs_show_enum *penum)
 
 /* Move to next character */
 private int
+show_fast_move(gs_state *pgs, gs_fixed_point *pwxy)
+{	int code = gx_path_add_rel_point_inline(pgs->path, pwxy->x, pwxy->y);
+	/* If the current position is out of range, don't try to move. */
+	if ( code == gs_error_limitcheck && pgs->clamp_coordinates )
+	  code = 0;
+	return code;
+}
+private int
 show_move(register gs_show_enum *penum)
 {	register gs_state *pgs = penum->pgs;
 	if ( penum->do_kern < 0 )
@@ -774,7 +793,7 @@ show_move(register gs_show_enum *penum)
 			gs_rmoveto(pgs, penum->wcx, penum->wcy);
 	}
 	/* wxy is in device coordinates */
-	   {	int code = show_fast_move(penum->wxy, pgs);
+	   {	int code = show_fast_move(pgs, &penum->wxy);
 		if ( code < 0 )
 		  return code;
 	   }
@@ -794,6 +813,10 @@ show_proceed(register gs_show_enum *penum)
 	gs_font *rfont = gs_rootfont(pgs);
 	int wmode = rfont->WMode;
 	font_proc_next_char((*next_char)) = rfont->procs.next_char;
+	font_proc_next_glyph((*next_glyph)) = rfont->procs.next_glyph;
+#define next_char_glyph(penum, pchr, pglyph)\
+  (next_char == 0 ? (*next_glyph)(penum, pchr, pglyph) :\
+   (*(pglyph) = gs_no_glyph, (*next_char)(penum, pchr)))
 	gs_char chr;
 	gs_glyph glyph;
 	int code;
@@ -802,7 +825,7 @@ show_proceed(register gs_show_enum *penum)
 	int alpha_bits = (*dev_proc(dev, get_alpha_bits))(dev, go_text);
 
 	if ( penum->charpath_flag == cpm_show && !penum->stringwidth_flag )
-	  {	code = gx_color_load(pgs->dev_color, pgs);
+	  {	code = gs_state_color_load(pgs);
 		if ( code < 0 )
 		  return code;
 	  }
@@ -814,7 +837,7 @@ more:	/* Proceed to next character */
 	if ( penum->can_cache >= 0 )
 	{	/* Loop with cache */
 		for ( ; ; )
-		{   switch ( (code = (*next_char)(penum, &chr)) )
+		{   switch ( (code = next_char_glyph(penum, &chr, &glyph)) )
 		    {
 		    default:			/* error */
 			return code;
@@ -834,11 +857,13 @@ more:	/* Proceed to next character */
 			 * do it here.
 			 */
 			penum->current_char = chr;
-			glyph = (*penum->encode_char)(penum, pfont, &chr);
-			penum->current_char = chr;
 			if ( glyph == gs_no_glyph )
-			  {	cc = 0;
+			  { glyph = (*penum->encode_char)(penum, pfont, &chr);
+			    penum->current_char = chr;
+			    if ( glyph == gs_no_glyph )
+			      { cc = 0;
 				goto no_cache;
+			      }
 			  }
 			if ( pair == 0 )
 			  pair = gx_lookup_fm_pair(pfont, pgs);
@@ -925,7 +950,7 @@ more:	/* Proceed to next character */
 			    code = show_move(penum);
 			  }
 			else
-			  code = show_fast_move(cc->wxy, pgs);
+			  code = show_fast_move(pgs, &cc->wxy);
 			if ( code )
 			  { /* Might be kshow, so store the state. */
 			    penum->current_glyph = glyph;
@@ -936,7 +961,7 @@ more:	/* Proceed to next character */
 	}
 	else
 	{	/* Can't use cache */
-		switch ( (code = (*next_char)(penum, &chr)) )
+		switch ( (code = next_char_glyph(penum, &chr, &glyph)) )
 		{
 		default:
 			return code;
@@ -949,8 +974,10 @@ more:	/* Proceed to next character */
 			;
 		}
 		penum->current_char = chr;
-		glyph = (*penum->encode_char)(penum, pfont, &chr);
-		penum->current_char = chr;
+		if ( glyph != gs_no_glyph )
+		  { glyph = (*penum->encode_char)(penum, pfont, &chr);
+		    penum->current_char = chr;
+		  }
 		cc = 0;
 	}
 no_cache:
@@ -972,15 +999,20 @@ no_cache:
 	pgs->in_cachedevice = 0;
 	/* Reset the sampling scale. */
 	penum->log2_current_scale.x = penum->log2_current_scale.y = 0;
-	/* Set the charpath flag in the graphics context if necessary, */
+	/* Set the charpath data in the graphics context if necessary, */
 	/* so that fill and stroke will add to the path */
 	/* rather than having their usual effect. */
 	pgs->in_charpath = penum->charpath_flag;
+	pgs->show_gstate =
+	  (penum->show_gstate == pgs ? pgs->saved : penum->show_gstate);
 	pgs->stroke_adjust = false;		/* per specification */
 	   {	gs_fixed_point cpt;
 		gx_path *ppath = pgs->path;
+
 		if ( (code = gx_path_current_point_inline(ppath, &cpt)) < 0 )
 		  goto rret;
+		penum->origin.x = cpt.x;
+		penum->origin.y = cpt.y;
 		/* Normally, char_tm is valid because of show_state_setup, */
 		/* but if we're in a cshow, it may not be. */
 		gs_currentcharmatrix(pgs, NULL, true);
@@ -1010,8 +1042,6 @@ no_cache:
 			cpt.x = float2fixed(fpx);
 			cpt.y = float2fixed(fpy);
 		  }
-		penum->origin.x = cpt.x;
-		penum->origin.y = cpt.y;
 		gs_newpath(pgs);
 		code = show_origin_setup(pgs, cpt.x, cpt.y,
 					 penum->charpath_flag);
@@ -1054,22 +1084,23 @@ no_cache:
 	/* we must undo the partial setup. */
 rret:	gs_grestore(pgs);
 	return code;
+#undef next_char_glyph
 }
 
 /* Finish show or stringwidth */
 private int
 show_finish(gs_show_enum *penum)
 {	gs_state *pgs = penum->pgs;
-	int code;
+	int code, rcode;
+
 	gs_show_enum_release(penum, NULL);
 	if ( penum->stringwidth_flag <= 0 )	/* could be cshow */
 	  return 0;
 	/* Save the accumulated width before returning, */
 	/* and undo the extra gsave. */
 	code = gs_currentpoint(pgs, &penum->width);
-	if ( code < 0 )
-	  return code;
-	return gs_grestore(pgs);
+	rcode = gs_grestore(pgs);
+	return (code < 0 ? code : rcode);
 }
 
 /* Return the current character for rendering. */
@@ -1144,7 +1175,8 @@ gs_show_width_only(const gs_show_enum *penum)
 /* Initialize a show enumerator. */
 /* Note that this does not set str.size. */
 private int
-show_setup(register gs_show_enum *penum, gs_state *pgs, const char *str)
+show_setup(register gs_show_enum *penum, gs_state *pgs, const char *str,
+  bool propagate_charpath)
 {	int code;
 	gs_font *pfont;
 	gx_set_dev_color(pgs);
@@ -1155,7 +1187,8 @@ show_setup(register gs_show_enum *penum, gs_state *pgs, const char *str)
 	penum->wchr = gs_no_char;
 	penum->add = false;
 	penum->do_kern = 0;
-	penum->charpath_flag = cpm_show;
+	penum->charpath_flag =
+	  (propagate_charpath ? pgs->in_charpath : cpm_show);
 	penum->stringwidth_flag = 0;
 	penum->dev_cache = 0;
 	penum->dev_cache2 = 0;
@@ -1166,11 +1199,14 @@ show_setup(register gs_show_enum *penum, gs_state *pgs, const char *str)
 	code = (*pfont->procs.init_fstack)(penum, pfont);
 	if ( code < 0 )
 	  return code;
-	penum->can_cache = 1;		/* show_state_setup may reset */
+	penum->can_cache =		/* show_state_setup may reset */
+	  (penum->charpath_flag == cpm_show ? 1 : -1);
 	code = show_state_setup(penum);
 	if ( code < 0 )
 	  return code;
-	pgs->show_gstate = pgs;
+	penum->show_gstate =
+	  (propagate_charpath && (pgs->in_charpath != 0) ?
+	   pgs->show_gstate : pgs);
 	return 0;
 }
 
@@ -1344,10 +1380,16 @@ gs_default_init_fstack(gs_show_enum *penum, gs_font *pfont)
 /* Default next-character procedure. */
 int
 gs_default_next_char(gs_show_enum *penum, gs_char *pchr)
+{	gs_glyph ignore_glyph;
+	return gs_default_next_glyph(penum, pchr, &ignore_glyph);
+}
+
+/* Default next-glyph procedure. */
+int
+gs_default_next_glyph(gs_show_enum *penum, gs_char *pchr, gs_glyph *pglyph)
 {	if ( penum->index == penum->str.size )
-		return 2;
-	else
-	{	*pchr = penum->str.data[penum->index++];
-		return 0;
-	}
+	  return 2;
+	*pchr = penum->str.data[penum->index++];
+	*pglyph = gs_no_glyph;
+	return 0;
 }
