@@ -1,24 +1,19 @@
-/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* gsline.c */
@@ -53,7 +48,9 @@ gs_currentlinewidth(const gs_state *pgs)
 /* setlinecap */
 int
 gs_setlinecap(gs_state *pgs, gs_line_cap cap)
-{	pgs_lp->cap = cap;
+{	if ( (uint)cap > gs_line_cap_max )
+	  return_error(gs_error_rangecheck);
+	pgs_lp->cap = cap;
 	return 0;
 }
 
@@ -66,7 +63,9 @@ gs_currentlinecap(const gs_state *pgs)
 /* setlinejoin */
 int
 gs_setlinejoin(gs_state *pgs, gs_line_join join)
-{	pgs_lp->join = join;
+{	if ( (uint)join > gs_line_join_max )
+	  return_error(gs_error_rangecheck);
+	pgs_lp->join = join;
 	return 0;
 }
 
@@ -131,11 +130,14 @@ gx_set_dash(gx_dash_params *dash, const float *pattern, uint length,
 	}
 	if ( length == 0 )		/* empty pattern */
 	{	dist_left = 0.0;
-		if ( mem )
-		  ppat = 0;
+		if ( mem && ppat )
+		  { gs_free_object(mem, ppat, "gx_set_dash(old pattern)");
+		    ppat = 0;
+		  }
 	}
 	else
-	{	if ( pattern_length == 0 )
+	{	uint size = length * sizeof(float);
+		if ( pattern_length == 0 )
 		  return_error(gs_error_rangecheck);
 		/* Compute the initial index, ink_on, and distance left */
 		/* in the pattern, according to the offset. */
@@ -155,9 +157,12 @@ gx_set_dash(gx_dash_params *dash, const float *pattern, uint length,
 		      )
 		  ink = !ink, index++;
 		if ( mem )
-		  { ppat = (float *)gs_alloc_bytes(mem,
-						   length * sizeof(float),
-						   "dash pattern");
+		  { if ( ppat == 0 )
+		      ppat = (float *)gs_alloc_bytes(mem, size,
+						     "gx_set_dash(pattern)");
+		    else if ( length != dash->pattern_size )
+		      ppat = gs_resize_object(mem, ppat, size,
+					      "gx_set_dash(pattern)");
 		    if ( ppat == 0 )
 		      return_error(gs_error_VMerror);
 		  }
@@ -166,6 +171,7 @@ gx_set_dash(gx_dash_params *dash, const float *pattern, uint length,
 	dash->pattern = ppat;
 	dash->pattern_size = length;
 	dash->offset = offset;
+	dash->pattern_length = pattern_length;
 	dash->init_ink_on = ink;
 	dash->init_index = index;
 	dash->init_dist_left = -dist_left;
@@ -232,3 +238,65 @@ bool
 gs_currentstrokeadjust(const gs_state *pgs)
 {	return pgs->stroke_adjust;
 }
+
+/* ------ Extensions ------ */
+
+/* Device-independent */
+
+/* setdashadapt */
+void
+gs_setdashadapt(gs_state *pgs, bool adapt)
+{	pgs_lp->dash.adapt = adapt;
+}
+
+/* currentdashadapt */
+bool
+gs_imager_currentdashadapt(const gs_imager_state *pis)
+{	return gs_currentlineparams_inline(pis)->dash.adapt;
+}
+bool
+gs_currentdashadapt(const gs_state *pgs)
+{	return gs_imager_currentdashadapt((const gs_imager_state *)pgs);
+}
+
+/* Device-dependent */
+
+/* setaccuratecurves */
+void
+gs_setaccuratecurves(gs_state *pgs, bool accurate)
+{	pgs->accurate_curves = accurate;
+}
+
+/* currentaccuratecurves */
+bool
+gs_imager_currentaccuratecurves(const gs_imager_state *pis)
+{	return pis->accurate_curves;
+}
+bool
+gs_currentaccuratecurves(const gs_state *pgs)
+{	return gs_imager_currentaccuratecurves((const gs_imager_state *)pgs);
+}
+
+/* setdotlength */
+int
+gx_set_dot_length(gx_line_params *plp, floatp length, bool absolute)
+{	if ( length < 0 )
+	  return_error(gs_error_rangecheck);
+	plp->dot_length = length;
+	plp->dot_length_absolute = absolute;
+	return 0;
+}
+int
+gs_setdotlength(gs_state *pgs, floatp length, bool absolute)
+{	return gx_set_dot_length(pgs_lp, length, absolute);
+}
+
+/* currentdotlength */
+float
+gs_currentdotlength(const gs_state *pgs)
+{	return pgs_lp->dot_length;
+}
+bool
+gs_currentdotlength_absolute(const gs_state *pgs)
+{	return pgs_lp->dot_length_absolute;
+}
