@@ -1,24 +1,19 @@
-/* Copyright (C) 1991, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1991, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* gxcpath.c */
@@ -29,9 +24,13 @@
 #include "gsutil.h"
 #include "gxdevice.h"
 #include "gxfixed.h"
+#include "gscoord.h"		/* needs gsmatrix.h */
 #include "gzpath.h"
 #include "gzcpath.h"
 
+/* Define whether to look for vertical clipping regions. */
+#define CHECK_VERTICAL_CLIPPING
+
 /* Imported from gxacpath.c */
 extern int gx_cpath_intersect_slow(P4(gs_state *, gx_clip_path *,
   gx_path *, int));
@@ -73,9 +72,8 @@ private ENUM_PTRS_BEGIN(device_clip_enum_ptrs) {
 					      index - (st_clip_list_max_ptrs + 1), pep);
 	}
 	case 0:
-		*pep = (cptr->current == &cptr->list.single ? NULL :
-			(void *)cptr->current);
-		break;
+	  ENUM_RETURN((cptr->current == &cptr->list.single ? NULL :
+		       (void *)cptr->current));
 ENUM_PTRS_END
 private RELOC_PTRS_BEGIN(device_clip_reloc_ptrs) {
 	if ( cptr->current == &cptr->list.single )
@@ -256,23 +254,44 @@ gx_cpath_intersect(gs_state *pgs, gx_clip_path *pcpath, gx_path *ppath,
   int rule)
 {	gs_fixed_rect old_box, new_box;
 	int code;
+
 	if ( gx_cpath_inner_box(pcpath, &old_box) &&
-	     gx_path_is_rectangle(ppath, &new_box)
+	     ((code = gx_path_is_rectangle(ppath, &new_box)) ||
+	      gx_path_is_void(ppath))
 	   )
 	   {	bool changed = false;
 		bool outside = pcpath->list.outside;
 
-		/* Intersect the two rectangles if necessary. */
-		if ( old_box.p.x > new_box.p.x )
-		  new_box.p.x = old_box.p.x, changed = true;
-		if ( old_box.p.y > new_box.p.y )
-		  new_box.p.y = old_box.p.y, changed = true;
-		if ( old_box.q.x < new_box.q.x )
-		  new_box.q.x = old_box.q.x, changed = true;
-		if ( old_box.q.y < new_box.q.y )
-		  new_box.q.y = old_box.q.y, changed = true;
-		if ( changed )
-		   {	/* Store the new rectangle back into the new path. */
+		if ( !code )
+		  { /* The new path is void. */
+		    if ( gx_path_current_point(ppath, &new_box.p) < 0 )
+		      { /* Use the user space origin (arbitrarily). */
+			gs_point origin;
+
+			gs_transform(pgs, 0.0, 0.0, &origin);
+			new_box.p.x = float2fixed(origin.x);
+			new_box.p.y = float2fixed(origin.y);
+		        gx_path_add_point(ppath, new_box.p.x, new_box.p.y);
+		      }
+		    new_box.q = new_box.p;
+		  }
+		else
+		  { /* Intersect the two rectangles if necessary. */
+		    if ( old_box.p.x > new_box.p.x )
+		      new_box.p.x = old_box.p.x, changed = true;
+		    if ( old_box.p.y > new_box.p.y )
+		      new_box.p.y = old_box.p.y, changed = true;
+		    if ( old_box.q.x < new_box.q.x )
+		      new_box.q.x = old_box.q.x, changed = true;
+		    if ( old_box.q.y < new_box.q.y )
+		      new_box.q.y = old_box.q.y, changed = true;
+		    /* Check for a degenerate rectangle. */
+		    if ( new_box.q.x < new_box.p.x )
+		      new_box.q.x = new_box.p.x;
+		    if ( new_box.q.y < new_box.p.y )
+		      new_box.q.y = new_box.p.y;
+		    if ( changed )
+		      { /* Store the new rectangle back into the new path. */
 			register segment *pseg =
 			  (segment *)ppath->first_subpath;
 #define set_pt(pqx,pqy)\
@@ -284,14 +303,17 @@ gx_cpath_intersect(gs_state *pgs, gx_clip_path *pcpath, gx_path *ppath,
 			if ( pseg != 0 ) /* might be an open rectangle */
 			  set_pt(p, p);
 #undef set_pt
-		   }
+		      }
+		  }
 		ppath->bbox = new_box;
+		gx_cpath_release(pcpath);
 		gx_clip_list_from_rectangle(&pcpath->list, &new_box);
 		pcpath->list.outside = outside;
 		pcpath->inner_box = new_box;
 		pcpath->path = *ppath;
 		gx_cpath_set_outer_box(pcpath);
 		pcpath->segments_valid = 1;
+		pcpath->shares_list = 0;
 		code = 1;
 		pcpath->id = gs_next_ids(1);	/* path changed => change id */
 	   }
@@ -361,24 +383,122 @@ gx_clip_list_from_rectangle(register gx_clip_list *clp,
 }
 
 /* Add a clip list to a path. */
-/* The current implementation is very inefficient. */
+/* In general, this produces a path made up of zillions of tiny lines. */
 private int
 gx_clip_list_add_to_path(gx_clip_list *clp, gx_path *ppath)
-{	const gx_clip_rect *rp;
+{	gx_clip_rect *rp;
 	int code = -1;
-	for ( rp = (clp->count <= 1 ? &clp->single : clp->head); rp != 0;
-	      rp = rp->next
-	    )
-	{	if ( rp->xmin < rp->xmax && rp->ymin < rp->ymax )
-		{	code = gx_path_add_rectangle(ppath,
-					int2fixed(rp->xmin),
-					int2fixed(rp->ymin),
-					int2fixed(rp->xmax),
-					int2fixed(rp->ymax));
-			if ( code < 0 )
-				return code;
+	gx_clip_rect *head = (clp->count <= 1 ? &clp->single : clp->head);
+	gx_clip_rect *visit;
+	gx_clip_rect *look;
+	enum { visit_left = 1, visit_right = 2 } first_visit;
+
+	for ( rp = head; rp != 0; rp = rp->next )
+	  if ( rp->xmin < rp->xmax && rp->ymin < rp->ymax )
+	    rp->to_visit = visit_left | visit_right;
+	for ( visit = head; visit != 0; visit = visit->next )
+	  { if ( !visit->to_visit )
+	      continue;
+	    rp = visit;
+	    if ( visit->to_visit & visit_left )
+	      { code = gx_path_add_point(ppath, int2fixed(visit->xmin),
+					 int2fixed(visit->ymax));
+	        if ( code < 0 )
+		  return code;
+		first_visit = visit_left;
+		goto left;
+	      }
+	    else
+	      { code = gx_path_add_point(ppath, int2fixed(visit->xmax),
+					 int2fixed(visit->ymin));
+	        if ( code < 0 )
+		  return code;
+		first_visit = visit_right;
+		goto right;
+	      }
+#define trace_line(px, py)\
+  code = gx_path_add_line(ppath, int2fixed(px), int2fixed(py));\
+  if ( code < 0 ) return code
+left:	    /* Trace upward along a left edge. */
+	    /* We're at the upper left corner of rp. */
+	    rp->to_visit &= ~visit_left;
+	    /* Look for an adjacent rectangle above rp. */
+	    for ( look = rp;
+		  (look = look->next) != 0 &&
+		    (look->ymin == rp->ymin ||
+		     (look->ymin == rp->ymax && look->xmax <= rp->xmin));
+		)
+	      ;
+	    /* Now we know look->ymin >= rp->ymax. */
+	    if ( look == 0 || look->ymin > rp->ymax || look->xmin >= rp->xmax )
+	      { /* No adjacent rectangle, switch directions. */
+		trace_line(rp->xmax, rp->ymax);
+		if ( rp == visit && first_visit == visit_right )
+		  goto close;
+		goto right1;
+	      }
+	    /* We found an adjacent rectangle. */
+	    /* See if it also adjoins a rectangle to the left of rp. */
+	    { gx_clip_rect *prev = rp->prev;
+	      if ( prev->ymax == rp->ymax && look->xmin < prev->xmax )
+		{ /* There's an adjoining rectangle as well. */
+		  /* Switch directions. */
+		  trace_line(prev->xmax, rp->ymax);
+		  rp = prev;
+		  if ( rp == visit && first_visit == visit_right )
+		    goto close;
+		  goto right1;
 		}
-	}
+	    }
+	    trace_line(look->xmin, look->ymin);
+	    rp = look;
+	    if ( rp == visit && first_visit == visit_left )
+	      goto close;
+left1:	    trace_line(rp->xmin, rp->ymax);
+	    goto left;
+right:	    /* Trace downward along a right edge. */
+	    /* We're at the lower right corner of rp. */
+	    rp->to_visit &= ~visit_right;
+	    /* Look for an adjacent rectangle below rp. */
+	    for ( look = rp;
+		  (look = look->prev) != 0 &&
+		    (look->ymax == rp->ymax ||
+		     (look->ymax == rp->ymin && look->xmin >= rp->xmax));
+		)
+	      ;
+	    /* Now we know look->ymax <= rp->ymin. */
+	    if ( look == 0 || look->ymax < rp->ymin || look->xmax <= rp->xmin )
+	      { /* No adjacent rectangle, switch directions. */
+		trace_line(rp->xmin, rp->ymin);
+		if ( rp == visit && first_visit == visit_left )
+		  goto close;
+		goto left1;
+	      }
+	    /* We found an adjacent rectangle. */
+	    /* See if it also adjoins a rectangle to the right of rp. */
+	    { gx_clip_rect *next = rp->next;
+	      if ( next->ymin == rp->ymin && look->xmax > next->xmin )
+		{ /* There's an adjoining rectangle as well. */
+		  /* Switch directions. */
+		  trace_line(next->xmin, rp->ymin);
+		  rp = next;
+		  if ( rp == visit && first_visit == visit_left )
+		    goto close;
+		  goto left1;
+		}
+	    }
+	    trace_line(look->xmax, look->ymax);
+	    rp = look;
+	    if ( rp == visit && first_visit == visit_right )
+	      goto close;
+right1:	    trace_line(rp->xmax, rp->ymin);
+	    goto right;
+close:	    /* We've gone all the way around an edge. */
+	    code = gx_path_close_subpath(ppath);
+	    if ( code < 0 )
+	      return code;
+	  }
+#undef trace_line
 	if ( code < 0 )
 	{	/* We didn't have any rectangles. */
 		code = gx_path_add_point(ppath, fixed_0, fixed_0);
@@ -412,6 +532,7 @@ private dev_proc_copy_alpha(clip_copy_alpha);
 private dev_proc_fill_mask(clip_fill_mask);
 private dev_proc_strip_tile_rectangle(clip_strip_tile_rectangle);
 private dev_proc_strip_copy_rop(clip_strip_copy_rop);
+private dev_proc_get_clipping_box(clip_get_clipping_box);
 
 /* The device descriptor. */
 private const gx_device_clip gs_clip_device =
@@ -452,7 +573,8 @@ private const gx_device_clip gs_clip_device =
 		gx_default_image_data,
 		gx_default_end_image,
 		clip_strip_tile_rectangle,
-		clip_strip_copy_rop
+		clip_strip_copy_rop,
+		clip_get_clipping_box
 	}
 };
 #define rdev ((gx_device_clip *)dev)
@@ -474,7 +596,7 @@ gx_make_clip_path_device(gx_device_clip *dev, const gx_clip_path *pcpath)
 /* Declare and initialize the cursor variables. */
 #ifdef DEBUG
 private ulong clip_loops, clip_in, clip_down, clip_up, clip_x, clip_no_x;
-private uint clip_interval = 100;
+private uint clip_interval = 10000;
 # define inc(v) v++
 # define print_clip()\
     if ( clip_loops % clip_interval == 0 )\
@@ -526,23 +648,57 @@ private uint clip_interval = 100;
  * the clipping region.  Usage:
  *	DO_CLIP(adjust for yc > yp if necessary,
  *		process(xc, yc, xec, yec) [must be an expression])
+ *
+ * Note that we look ahead to detect unclipped vertical strips.
+ * This is really only valuable for 90 degree rotated images or
+ * (nearly-)vertical lines with convex clipping regions; if we ever
+ * change images to use source buffering and destination-oriented
+ * enumeration, we could probably take out the code here with no
+ * adverse effects.
  */
+#ifdef CHECK_VERTICAL_CLIPPING
+#  define LOOK_AHEAD\
+				if ( xec - xc == w )	/* full width */\
+				  { /* Look ahead for a vertical swath. */\
+				    while ( (nptr = rptr->next) != 0 &&\
+					    nptr->ymin == yec &&\
+					    nptr->ymax <= ye &&\
+					    nptr->xmin <= x &&\
+					    nptr->xmax >= xe\
+					  )\
+				      yec = nptr->ymax, rptr = nptr;\
+				  }\
+				else\
+				  nptr = rptr->next
+#else
+#  define LOOK_AHEAD\
+				  nptr = rptr->next
+#endif
 #define DO_CLIP(adjust_for_y, process_rectangle)\
 	if ( w <= 0 || h <= 0 ) return 0;\
 	inc(clip_loops);\
+	print_clip();\
    {	const int xe = x + w, ye = y + h;\
 	int xc, xec, yc, yec, yp, yep;\
 	int code;\
+\
 	warp_cursor(y);\
 	if ( rptr == 0 || (yc = rptr->ymin) >= ye )\
-	  return (outside ? (xc = x, xec = xe, yc = y, yec = ye,\
-			     process_rectangle) : 0);\
+	  { if ( rdev->list.count > 1 )\
+	      rdev->current =\
+		(rptr != 0 ? rptr :\
+		 y >= rdev->current->ymax ? rdev->list.tail :\
+		 rdev->list.head);\
+	    return (outside ? (xc = x, xec = xe, yc = y, yec = ye,\
+			       process_rectangle) : 0);\
+	  }\
 	rdev->current = rptr;\
 	if ( yc < y ) yc = y;\
 	yp = y;\
 	if ( outside )\
 	  { for ( yep = y; ; )\
 	      { const int ymax = rptr->ymax;\
+\
 		xc = x;\
 		if ( yc > yep )\
 		  { yec = yc, yc = yep;\
@@ -564,6 +720,10 @@ private uint clip_interval = 100;
 				xc = rptr->xmax;\
 				if ( xc >= xe ) xc = max_int;\
 			   }\
+			else\
+			  { xec = rptr->xmax;\
+			    if ( xec > xc ) xc = xec;\
+			  }\
 		   }\
 		while ( (rptr = rptr->next) != 0 && rptr->ymax == ymax );\
 		if ( xc < xe )\
@@ -584,6 +744,8 @@ private uint clip_interval = 100;
 	else \
 	  for ( ; ; )\
 	    {	const int ymax = rptr->ymax;\
+		gx_clip_rect *nptr;\
+\
 		yec = min(ymax, ye);\
 		if ( yc > yp ) adjust_for_y;\
 		if_debug2('Q', "[Q]yc=%d yec=%d\n", yc, yec);\
@@ -596,17 +758,20 @@ private uint clip_interval = 100;
 			   {	clip_rect_print('Q', "match", rptr);\
 				if_debug2('Q', "[Q]xc=%d xec=%d\n", xc, xec);\
 				inc(clip_x);\
+				LOOK_AHEAD;\
 				code = process_rectangle;\
 				if ( code < 0 ) return code;\
 			   }\
-			else inc(clip_no_x);\
+			else\
+			  { inc(clip_no_x);\
+			    nptr = rptr->next;\
+			  }\
 		   }\
-		while ( (rptr = rptr->next) != 0 && rptr->ymax == ymax );\
+		while ( (rptr = nptr) != 0 && rptr->ymax == ymax );\
 		if ( rptr == 0 || (yec = rptr->ymin) >= ye ) break;\
 		yp = yc;\
 		yc = yec;\
 	    }\
-	print_clip();\
    }
 
 /* Open a clipping device */
@@ -617,6 +782,8 @@ clip_open(register gx_device *dev)
 	rdev->current =
 	  (rdev->list.head == 0 ? &rdev->list.single : rdev->list.head);
 	rdev->color_info = tdev->color_info;
+	rdev->width = tdev->width;
+	rdev->height = tdev->height;
 	return 0;
 }
 
@@ -763,6 +930,52 @@ clip_strip_copy_rop(gx_device *dev,
 	return 0;
 }
 
+/* Get the (outer) clipping box, in client coordinates. */
+private void
+clip_get_clipping_box(gx_device *dev, gs_fixed_rect *pbox)
+{	gx_device *tdev = rdev->target;
+	gs_fixed_rect tbox, cbox;
+	fixed tx = int2fixed(rdev->translation.x),
+	  ty = int2fixed(rdev->translation.y);
+
+	(*dev_proc(tdev, get_clipping_box))(tdev, &tbox);
+	/*
+	 * To get an accurate clipping box quickly in all cases, we should
+	 * save the outer box from the clipping path.  However,
+	 * this is not currently (or even always guaranteed to be)
+	 * available.  Instead, we compromise: if there is more than one
+	 * rectangle in the list, we return accurate Y values (which are
+	 * easy to obtain, because the list is Y-sorted) but copy the
+	 * X values from the target.
+	 */
+	if ( rdev->list.outside || rdev->list.count == 0 )
+	  { cbox = tbox;
+	  }
+	else if ( rdev->list.count == 1 )
+	  { cbox.p.x = int2fixed(rdev->list.single.xmin);
+	    cbox.p.y = int2fixed(rdev->list.single.ymin);
+	    cbox.q.x = int2fixed(rdev->list.single.xmax);
+	    cbox.q.y = int2fixed(rdev->list.single.ymax);
+	  }
+	else
+	  { /* The head and tail elements are dummies.... */
+	    cbox.p.x = tbox.p.x;
+	    cbox.p.y = int2fixed(rdev->list.head->next->ymin);
+	    cbox.q.x = tbox.q.x;
+	    cbox.q.y = int2fixed(rdev->list.tail->prev->ymax);
+	  }
+	rect_intersect(tbox, cbox);
+	if ( tbox.p.x != min_fixed )
+	  tbox.p.x -= tx;
+	if ( tbox.p.y != min_fixed )
+	  tbox.p.y -= ty;
+	if ( tbox.q.x != max_fixed )
+	  tbox.q.x -= tx;
+	if ( tbox.q.y != max_fixed )
+	  tbox.q.y -= ty;
+	*pbox = tbox;
+}
+
 /* ------ Debugging printout ------ */
 
 #ifdef DEBUG
