@@ -1,24 +1,19 @@
-/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* gxfill.c */
@@ -176,12 +171,12 @@ private int add_y_list(P5(gx_path *, ll_ptr, fixed, fixed,
 private int add_y_line(P4(const segment *, const segment *, int, ll_ptr));
 private void near insert_x_new(P2(active_line *, ll_ptr));
 private bool near end_x_line(P1(active_line *));
-private int fill_loop_by_scan_lines(P11(ll_ptr, gx_device *,
-  const gx_fill_params *, const gx_device_color *, gs_logical_operation_t,
-  const gs_fixed_rect *, fixed, fixed, fixed, fixed, fixed));
-private int fill_loop_by_trapezoids(P11(ll_ptr, gx_device *,
-  const gx_fill_params *, const gx_device_color *, gs_logical_operation_t,
-  const gs_fixed_rect *, fixed, fixed, fixed, fixed, fixed));
+#define fill_loop_proc(proc)\
+int proc(P11(ll_ptr, gx_device *,\
+  const gx_fill_params *, const gx_device_color *, gs_logical_operation_t,\
+  const gs_fixed_rect *, fixed, fixed, fixed, fixed, fixed))
+private fill_loop_proc(fill_loop_by_scan_lines);
+private fill_loop_proc(fill_loop_by_trapezoids);
 
 /* Statistics */
 #ifdef DEBUG
@@ -262,7 +257,7 @@ gx_default_fill_path(gx_device *pdev, const gs_imager_state *pis,
 #define adjust_x adjust.x
 #define adjust_y adjust.y
 	gs_logical_operation_t lop = pis->log_op;
-	gs_fixed_rect ibox, cbox;
+	gs_fixed_rect ibox, bbox;
 	gx_device_clip cdev;
 	gx_device *dev = pdev;
 	gx_device *save_dev = dev;
@@ -292,14 +287,17 @@ gx_default_fill_path(gx_device *pdev, const gs_imager_state *pis,
 		  fixed2float(adjust_x), fixed2float(adjust_y),
 		  fixed2float(ibox.p.x), fixed2float(ibox.p.y),
 		  fixed2float(ibox.q.x), fixed2float(ibox.q.y));
-	if ( pcpath != NULL &&
-	     (gx_cpath_inner_box(pcpath, &cbox), !rect_within(ibox, cbox))
-	   )
-	   {	/* Intersect the path box and the clip bounding box. */
-		/* If the intersection is empty, this fill is a no-op. */
-		gs_fixed_rect bbox;
-
-		gx_cpath_outer_box(pcpath, &bbox);
+	if ( pcpath )
+	  gx_cpath_inner_box(pcpath, &bbox);
+	else
+	  (*dev_proc(dev, get_clipping_box))(dev, &bbox);
+	if ( !rect_within(ibox, bbox) )
+	  {	/*
+		 * Intersect the path box and the clip bounding box.
+		 * If the intersection is empty, this fill is a no-op.
+		 */
+		if ( pcpath )
+		  gx_cpath_outer_box(pcpath, &bbox);
 		if_debug4('f', "   outer_box=(%g,%g),(%g,%g)\n",
 			  fixed2float(bbox.p.x), fixed2float(bbox.p.y),
 			  fixed2float(bbox.q.x), fixed2float(bbox.q.y));
@@ -319,27 +317,35 @@ gx_default_fill_path(gx_device *pdev, const gs_imager_state *pis,
 		 * recompute its bbox and make the tests again,
 		 * but we don't bother right now.
 		 *
-		 * Set up a clipping device.
+		 * If there is a clipping path, set up a clipping device.
 		 */
-		dev = (gx_device *)&cdev;
-		gx_make_clip_device(&cdev, &cdev, &pcpath->list);
-		cdev.target = save_dev;
-		cdev.max_fill_band = save_dev->max_fill_band;
-		(*dev_proc(dev, open_device))(dev);
-	   }
+		if ( pcpath )
+		  { dev = (gx_device *)&cdev;
+		    gx_make_clip_device(&cdev, &cdev, &pcpath->list);
+		    cdev.target = save_dev;
+		    cdev.max_fill_band = save_dev->max_fill_band;
+		    (*dev_proc(dev, open_device))(dev);
+		  }
+	  }
 	/*
 	 * Compute the proper adjustment values.
 	 * To get the effect of the any-part-of-pixel rule,
 	 * we may have to tweak them slightly.
+	 * NOTE: We changed the adjust_right/above value from 0.5+epsilon
+	 * to 0.5 in release 5.01; even though this does the right thing
+	 * in every case we could imagine, we aren't confident that it's
+	 * correct.  (The old values were definitely incorrect, since they
+	 * caused 1-pixel-wide/high objects to color 2 pixels even if
+	 * they fell exactly on pixel boundaries.)
 	 */
 	if ( adjust.x == fixed_half )
 	  adjust_left = fixed_half - fixed_epsilon,
-	  adjust_right = fixed_half + fixed_epsilon;
+	  adjust_right = fixed_half /* + fixed_epsilon */;  /* see above */
 	else
 	  adjust_left = adjust_right = adjust.x;
 	if ( adjust.y == fixed_half )
 	  adjust_below = fixed_half - fixed_epsilon,
-	  adjust_above = fixed_half + fixed_epsilon;
+	  adjust_above = fixed_half /* + fixed_epsilon */;  /* see above */
 	else
 	  adjust_below = adjust_above = adjust.y;
 	/* Initialize the active line list. */
@@ -365,8 +371,8 @@ gx_default_fill_path(gx_device *pdev, const gs_imager_state *pis,
 #ifdef FILL_SCAN_LINES
 #  ifdef FILL_TRAPEZOIDS
 	fill_by_trapezoids =
-	  ((adjust_below | adjust_above) != 0 || ppath->curve_count == 0) &&
-	    double_write_ok;
+	  ((adjust_below | adjust_above) != 0 || ppath->curve_count == 0 ||
+	   params->flatness >= 1.0) && double_write_ok;
 #  else
 	fill_by_trapezoids = false;
 #  endif
@@ -384,7 +390,9 @@ gx_default_fill_path(gx_device *pdev, const gs_imager_state *pis,
 	else
 #ifdef FILL_CURVES
 	if ( fill_by_trapezoids )
-	  {	code = gx_path_flatten(ppath, &ffpath, params->flatness);
+	  {	code = gx_path_flatten_accurate(ppath, &ffpath,
+						params->flatness,
+						pis->accurate_curves);
 		if ( code < 0 )
 		  return code;
 		pfpath = &ffpath;
@@ -398,7 +406,9 @@ gx_default_fill_path(gx_device *pdev, const gs_imager_state *pis,
 		pfpath = &ffpath;
 	  }
 #else
-	  {	code = gx_path_flatten(ppath, &ffpath, params->flatness);
+	  {	code = gx_path_flatten_accurate(ppath, &ffpath,
+						params->flatness,
+						pis->accurate_curves);
 		if ( code < 0 )
 		  return code;
 		pfpath = &ffpath;
@@ -406,12 +416,18 @@ gx_default_fill_path(gx_device *pdev, const gs_imager_state *pis,
 #endif
 	if ( (code = add_y_list(pfpath, &lst, adjust_below, adjust_above, &ibox)) < 0 )
 	  goto nope;
-	code =
-	  (fill_by_trapezoids ? fill_loop_by_trapezoids :
-	   fill_loop_by_scan_lines)
+	{ fill_loop_proc((*fill_loop));
+
+	  /* Some short-sighted compilers won't allow a conditional here.... */
+	  if ( fill_by_trapezoids )
+	    fill_loop = fill_loop_by_trapezoids;
+	  else
+	    fill_loop = fill_loop_by_scan_lines;
+	  code = (*fill_loop)
 	    (&lst, dev, params, pdevc, lop, &ibox,
 	     adjust_left, adjust_right, adjust_below, adjust_above,
 	     (max_fill_band == 0 ? no_band_mask : int2fixed(-max_fill_band)));
+	}
 nope:	if ( lst.close_count != 0 )
 	  unclose_path(pfpath, lst.close_count);
 	free_line_list(&lst);
@@ -968,27 +984,31 @@ set_scan_line_points(active_line *alp, fixed fixed_flat)
 /* ---------------- Trapezoid filling loop ---------------- */
 
 /* Forward references */
-private int near fill_slant_adjust(P11(fixed, fixed, fixed, fixed, fixed,
-  fixed, fixed, fixed, const gx_device_color *, gx_device *,
-  gs_logical_operation_t));
+private int near fill_slant_adjust(P12(fixed, fixed, fixed, fixed, fixed,
+  fixed, fixed, fixed, const gs_fixed_rect *,
+  const gx_device_color *, gx_device *, gs_logical_operation_t));
 private void near resort_x_line(P1(active_line *));
 
 /****** PATCH ******/
-#define loop_fill_trapezoid_fixed(fx0, fw0, fy0, fx1, fw1, fh, swap_axes)\
-  loop_fill_trap(dev, fx0, fw0, fy0, fx1, fw1, fh, swap_axes, pdevc, lop)
+#define loop_fill_trapezoid_fixed(fx0, fw0, fy0, fx1, fw1, fh)\
+  loop_fill_trap(dev, fx0, fw0, fy0, fx1, fw1, fh, pbox, pdevc, lop)
 private int
 loop_fill_trap(gx_device *dev, fixed fx0, fixed fw0, fixed fy0,
-  fixed fx1, fixed fw1, fixed fh, bool swap_axes,
+  fixed fx1, fixed fw1, fixed fh, const gs_fixed_rect *pbox,
   const gx_device_color *pdevc, gs_logical_operation_t lop)
 {	fixed fy1 = fy0 + fh;
+	fixed ybot = max(fy0, pbox->p.y);
+	fixed ytop = min(fy1, pbox->q.y);
 	gs_fixed_edge left, right;
 
+	if ( ybot >= ytop )
+	  return 0;
 	left.start.y = right.start.y = fy0;
 	left.end.y = right.end.y = fy1;
 	right.start.x = (left.start.x = fx0) + fw0;
 	right.end.x = (left.end.x = fx1) + fw1;
 	return (*dev_proc(dev, fill_trapezoid))
-	  (dev, &left, &right, fy0, fy1, swap_axes, pdevc, lop);
+	  (dev, &left, &right, ybot, ytop, false, pdevc, lop);
 }
 /****** END PATCH ******/
 
@@ -1002,7 +1022,7 @@ fill_loop_by_trapezoids(ll_ptr ll, gx_device *dev,
   fixed adjust_left, fixed adjust_right,
   fixed adjust_below, fixed adjust_above, fixed band_mask)
 {	int rule = params->rule;
-	fixed y_limit = pbox->q.y;
+	const fixed y_limit = pbox->q.y;
 	active_line *yll = ll->y_list;
 	fixed y;
 	int code;
@@ -1015,24 +1035,27 @@ fill_loop_by_trapezoids(ll_ptr ll, gx_device *dev,
  *	fixed2int_pixround(y - below) != fixed2int_pixround(y + above)
  * where we know
  *	0 <= below <= _fixed_pixround_v,
- *	0 <= above <= fixed_1 - below.
+ *	0 <= above <= min(fixed_half, fixed_1 - below).
  * Subtracting out the integer parts, this is equivalent to
  *	fixed2int_pixround(fixed_fraction(y) - below) !=
  *	  fixed2int_pixround(fixed_fraction(y) + above)
  * or to
  *	fixed2int(fixed_fraction(y) + _fixed_pixround_v - below) !=
  *	  fixed2int(fixed_fraction(y) + _fixed_pixround_v + above)
+ * Letting A = _fixed_pixround_v - below and B = _fixed_pixround_v + above,
+ * we can rewrite this as
+ *	fixed2int(fixed_fraction(y) + A) != fixed2int(fixed_fraction(y) + B)
  * Because of the range constraints given above, this is true precisely when
- *	fixed_fraction(y) + _fixed_pixround_v - below < fixed_1 &&
- *	  fixed_fraction(y) + _fixed_pixround_v + above >= fixed_1
- * or equivalently when
- *	fixed_fraction(y - (fixed_1 - _fixed_pixround_v - below)) <
- *	  below + above
+ *	fixed_fraction(y) + A < fixed_1 && fixed_fraction(y) + B >= fixed_1
+ * or equivalently
+ *	fixed_fraction(y + B) < B - A.
+ * i.e.
+ *	fixed_fraction(y + _fixed_pixround_v + above) < below + above
  */
-	fixed half_minus_adjust_y = fixed_1 - _fixed_pixround_v - adjust_below;
-	fixed adjust_y2 = adjust_below + adjust_above;
+	fixed y_span_delta = _fixed_pixround_v + adjust_above;
+	fixed y_span_limit = adjust_below + adjust_above;
 #define adjusted_y_spans_pixel(y)\
-  fixed_fraction((y) - half_minus_adjust_y) < adjust_y2
+  fixed_fraction((y) + y_span_delta) < y_span_limit
 
 	if ( yll == 0 ) return 0;		/* empty list */
 	if ( fill_direct )
@@ -1260,7 +1283,7 @@ fill_loop_by_trapezoids(ll_ptr ll, gx_device *dev,
 			     {	/* Top wider than bottom. */
 				code = loop_fill_trapezoid_fixed(
 					xlbot, wbot, y - adjust_below,
-					xltop, wtop, height, false);
+					xltop, wtop, height);
 				if ( adjusted_y_spans_pixel(y1) )
 				   {	if ( code < 0 ) return code;
 					n_add1(n_afill);
@@ -1273,7 +1296,8 @@ fill_loop_by_trapezoids(ll_ptr ll, gx_device *dev,
 			    {	/* Slanted trapezoid. */
 				code = fill_slant_adjust(xlbot, xbot, y,
 					xltop, xtop, height, adjust_below,
-					adjust_above, pdevc, dev, lop);
+					adjust_above, pbox,
+					pdevc, dev, lop);
 			    }
 			 }
 			else
@@ -1289,20 +1313,20 @@ fill_loop_by_trapezoids(ll_ptr ll, gx_device *dev,
 				   }
 				code = loop_fill_trapezoid_fixed(
 					xlbot, wbot, y + adjust_above,
-					xltop, wtop, height, false);
+					xltop, wtop, height);
 			    }
 			   else
 			    {	/* Slanted trapezoid. */
 				code = fill_slant_adjust(xlbot, xbot, y,
 					xltop, xtop, height, adjust_below,
-					adjust_above, pdevc, dev, lop);
+					adjust_above, pbox,
+					pdevc, dev, lop);
 			    }
 			 }
 		}
 		else		/* No Y adjustment. */
 		  code = loop_fill_trapezoid_fixed(xlbot, xbot - xlbot,
-						   y, xltop, wtop, height,
-						   false);
+						   y, xltop, wtop, height);
 		if ( code < 0 ) return code;
 	   }
 
@@ -1345,127 +1369,117 @@ fill_loop_by_trapezoids(ll_ptr ll, gx_device *dev,
 	return 0;
 }
 
-/* Handle the case of a slanted trapezoid with adjustment. */
-/* To do this exactly right requires filling a central trapezoid */
-/* plus two narrow vertical triangles or two horizontal almost-rectangles. */
-/* Yuck! */
+/*
+ * Handle the case of a slanted trapezoid with adjustment.
+ * To do this exactly right requires filling a central trapezoid
+ * (or rectangle) plus two horizontal almost-rectangles.
+ */
 private int near
 fill_slant_adjust(fixed xlbot, fixed xbot, fixed y,
   fixed xltop, fixed xtop, fixed height, fixed adjust_below,
-  fixed adjust_above, const gx_device_color *pdevc, gx_device *dev,
+  fixed adjust_above, const gs_fixed_rect *pbox,
+  const gx_device_color *pdevc, gx_device *dev,
   gs_logical_operation_t lop)
-{	fixed adjust_y2 = adjust_below + adjust_above;
-	fixed y1 = y + height;
+{	fixed y1 = y + height;
 	dev_proc_fill_trapezoid((*fill_trap)) =
 	  dev_proc(dev, fill_trapezoid);
+	const fixed yb = y - adjust_below;
+	const fixed ya = y + adjust_above;
+	const fixed y1b = y1 - adjust_below;
+	const fixed y1a = y1 + adjust_above;
+	const gs_fixed_edge *plbot;
+	const gs_fixed_edge *prbot;
+	const gs_fixed_edge *pltop;
+	const gs_fixed_edge *prtop;
+	gs_fixed_edge vert_left, slant_left, vert_right, slant_right;
 	int code;
 
 	n_add1(n_slant);
-	if ( height < adjust_y2 )
+
+	/* Set up all the edges, even though we may not need them all. */
+
+	if ( xlbot < xltop )	/* && xbot < xtop */
+	  { vert_left.start.x = vert_left.end.x = xlbot;
+	    vert_left.start.y = yb, vert_left.end.y = ya;
+	    vert_right.start.x = vert_right.end.x = xtop;
+	    vert_right.start.y = y1b, vert_right.end.y = y1a;
+	    slant_left.start.y = ya, slant_left.end.y = y1a;
+	    slant_right.start.y = yb, slant_right.end.y = y1b;
+	    plbot = &vert_left, prbot = &slant_right,
+	      pltop = &slant_left, prtop = &vert_right;
+	  }
+	else
+	  { vert_left.start.x = vert_left.end.x = xltop;
+	    vert_left.start.y = y1b, vert_left.end.y = y1a;
+	    vert_right.start.x = vert_right.end.x = xbot;
+	    vert_right.start.y = yb, vert_right.end.y = ya;
+	    slant_left.start.y = yb, slant_left.end.y = y1b;
+	    slant_right.start.y = ya, slant_right.end.y = y1a;
+	    plbot = &slant_left, prbot = &vert_right,
+	      pltop = &vert_left, prtop = &slant_right;
+	  }
+	slant_left.start.x = xlbot, slant_left.end.x = xltop;
+	slant_right.start.x = xbot, slant_right.end.x = xtop;
+
+	if ( ya >= y1b )
 	  {	/*
 		 * The upper and lower adjustment bands overlap.
-		 * Use a different algorithm.
 		 * Since the entire entity is less than 2 pixels high
 		 * in this case, we could handle it very efficiently
 		 * with no more than 2 rectangle fills, but for right now
-		 * we won't attempt to do this.
+		 * we don't attempt to do this.
 		 */
-		fixed xl, wx;
-		int yi, hi;
+		int iyb = fixed2int_var_pixround(yb);
+		int iya = fixed2int_var_pixround(ya);
+		int iy1b = fixed2int_var_pixround(y1b);
+		int iy1a = fixed2int_var_pixround(y1a);
 
 		n_add1(n_slant_shallow);
-		if ( xltop >= xlbot )		/* && xtop >= xbot */
-		  wx = xtop - (xl = xlbot);
-		else
-		  wx = xbot - (xl = xltop);
-		code = loop_fill_trapezoid_fixed(xlbot, xbot - xlbot,
-						 y - adjust_below,
-						 xl, wx, height, false);
-		if ( code < 0 )
-		  return code;
-		yi = fixed2int_pixround(y1 - adjust_below);
-		hi = fixed2int_pixround(y + adjust_above) - yi;
-		if ( hi > 0 )
-		  {	int xi = fixed2int_var_pixround(xl);
-			int wi = fixed2int_pixround(xl + wx) - xi;
-			code = loop_fill_rectangle(xi, yi, wi, hi);
-			if ( code < 0 )
-			  return code;
+		if ( iy1b > iyb )
+		  { code = (*fill_trap)(dev, plbot, prbot,
+					yb, y1b, false, pdevc, lop);
+		    if ( code < 0 )
+		      return code;
 		  }
-		code = loop_fill_trapezoid_fixed(xl, wx, y + adjust_above,
-						 xltop, xtop - xltop, height,
-						 false);
+		if ( iya > iy1b )
+		  { int ix = fixed2int_var_pixround(vert_left.start.x);
+		    int iw = fixed2int_var_pixround(vert_right.start.x) - ix;
+		    code = loop_fill_rectangle(ix, iy1b, iw, iya - iy1b);
+		    if ( code < 0 )
+		      return code;
+		  }
+		if ( iy1a > iya )
+		  code = (*fill_trap)(dev, pltop, prtop,
+				      ya, y1a, false, pdevc, lop);
+		else
+		  code = 0;
 	  }
 	else
-	  {	fixed dx_left = xltop - xlbot, dx_right = xtop - xbot;
-		fixed xlb, xrb, xlt, xrt;
-		int xli, xri;
-		fixed half_minus_adjust_y =
-		  fixed_1 - _fixed_pixround_v - adjust_below;
-
-		if ( dx_left <= 0 )		/* && dx_right <= 0 */
-		  {	xlb = xlbot -
-			  fixed_mult_quo(-dx_left, adjust_y2, height);
-			xrb = xbot;
-			xlt = xltop;
-			xrt = xtop +
-			  fixed_mult_quo(-dx_right, adjust_y2, height);
+	  {	/*
+		 * Clip the trapezoid if possible.  This can save a lot
+		 * of work when filling paths that cross band boundaries.
+		 */
+		fixed yac;
+	        if ( pbox->p.y < ya )
+		  { code = (*fill_trap)(dev, plbot, prbot,
+					yb, ya, false, pdevc, lop);
+		    if ( code < 0 )
+		      return code;
+		    yac = ya;
 		  }
-		else		/* dx_left >= 0, dx_right >= 0 */
-		  {	xlb = xlbot;
-			xrb = xbot +
-			  fixed_mult_quo(dx_right, adjust_y2, height);
-			xlt = xltop -
-			  fixed_mult_quo(dx_left, adjust_y2, height);
-			xrt = xtop;
-		  }
-		/* Do the bottom adjustment band, if any. */
-		if ( adjusted_y_spans_pixel(y) )
-		  {	/* We can always do this with a rectangle, */
-			/* but the computation may be too much trouble. */
-			n_add1(n_sfill);
-			if ( (xli = fixed2int_var_pixround(xlbot)) ==
-			      fixed2int_var_pixround(xlb) &&
-			     (xri = fixed2int_var_pixround(xbot)) ==
-			      fixed2int_var_pixround(xrb)
-			   )
-			  code =
-			    loop_fill_rectangle(xli, fixed2int_pixround(y - adjust_below),
-						xri - xli, 1);
-			else
-			  code =
-			    loop_fill_trapezoid_fixed(xlbot, xbot - xlbot,
-						      y - adjust_below,
-						      xlb, xrb - xlb,
-						      adjust_y2, false);
-			if ( code < 0 )
-			  return code;
-		  }
-		/* Do the central trapezoid. */
-		code = loop_fill_trapezoid_fixed(xlb, xrb - xlb,
-						 y + adjust_above,
-						 xlt, xrt - xlt,
-						 height - adjust_y2, false);
-		/* Do the top adjustment band. */
-		if ( adjusted_y_spans_pixel(y1) )
-		  {	if ( code < 0 )
-			  return code;
-			n_add1(n_sfill);
-			if ( (xli = fixed2int_var_pixround(xltop)) ==
-			      fixed2int_var_pixround(xlt) &&
-			     (xri = fixed2int_var_pixround(xtop)) ==
-			      fixed2int_var_pixround(xrt)
-			   )
-			  code =
-			    loop_fill_rectangle(xli, fixed2int_pixround(y1 - adjust_below),
-						xri - xli, 1);
-			else
-			  code =
-			    loop_fill_trapezoid_fixed(xlt, xrt - xlt,
-						      y1 - adjust_below,
-						      xltop, xtop - xltop,
-						      adjust_y2, false);
+		else
+		  yac = pbox->p.y;
+		if ( pbox->q.y > y1b )
+		  { code = (*fill_trap)(dev, &slant_left, &slant_right,
+					yac, y1b, false, pdevc, lop);
+		    if ( code < 0 )
+		      return code;
+		    code = (*fill_trap)(dev, pltop, prtop,
+					y1b, y1a, false, pdevc, lop);
 		  }
+		else
+		  code = (*fill_trap)(dev, &slant_left, &slant_right,
+				      yac, pbox->q.y, false, pdevc, lop);
 	  }
 	return code;
 }
