@@ -1,24 +1,19 @@
-/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* gxstroke.c */
@@ -29,9 +24,10 @@
 #include "gserrors.h"
 #include "gsdcolor.h"
 #include "gxfixed.h"
-#include "gxarith.h"
+#include "gxfarith.h"
 #include "gxmatrix.h"
 #include "gscoord.h"
+#include "gsdevice.h"
 #include "gxdevice.h"
 #include "gxhttile.h"
 #include "gxistate.h"
@@ -56,16 +52,43 @@
 #  define stroke_adjustment(thin, pis, xy) fixed_0
 #endif
 
+/*
+ * For some reason, we commented out the optimization for portrait,
+ * landscape, and uniform (non-scaled) transformations.  We have no record
+ * of why we did this, and we don't know what bugs re-enabling it may
+ * introduce.
+ */
+#define OPTIMIZE_ORIENTATION
+
 /*
  * Compute the amount by which to expand a stroked bounding box to account
- * for line width, caps and joins.  Because of square caps and miter joins,
- * the maximum expansion on each side is
- *	max(sqrt(2), miter_limit) * line_width/2,
- * computed in user space.
+ * for line width, caps and joins.  Because of square caps and miter and
+ * triangular joins, the maximum expansion on each side (in user space) is
+ *	K * line_width/2
+ * where K is determined as follows:
+ *	If the path is only a single line segment, K = 1;
+ *	if triangular joins, K = 2;
+ *	if miter joins, K = miter_limit;
+ *	otherwise, K = 1.
+ * If the amount is too large to fit in a gs_fixed_point, return
+ * gs_error_limitcheck.
+ *
+ * If the miter limit is very large, the foregoing computation will produce
+ * a result that is much too large; we would like to tighten this up at
+ * some point in the future.
  */
 int
-gx_stroke_expansion(const gs_imager_state *pis, gs_fixed_point *ppt)
-{	double expand = max(1.415, pis->line_params.miter_limit) *
+gx_stroke_path_expansion(const gs_imager_state *pis, const gx_path *ppath,
+  gs_fixed_point *ppt)
+{	const subpath *psub = ppath->first_subpath;
+	const segment *pseg;
+	double expand =
+	  (ppath->curve_count == 0 && ppath->subpath_count <= 1 &&
+	   (psub == 0 || (pseg = psub->next) == 0 ||
+	    (pseg = pseg->next) == 0 || pseg->type == s_line_close) ? 1.0 :
+	   pis->line_params.join == gs_join_miter ?
+	     pis->line_params.miter_limit :
+	   pis->line_params.join == gs_join_triangle ? 2.0 : 1.0) *
 	  pis->line_params.half_width;
 	/* Short-cut gs_bbox_transform. */
 	float cx1 = pis->ctm.xx + pis->ctm.yx;
@@ -77,9 +100,14 @@ gx_stroke_expansion(const gs_imager_state *pis, gs_fixed_point *ppt)
 	if ( cy1 < 0 ) cy1 = -cy1;
 	if ( cx2 < 0 ) cx2 = -cx2;
 	if ( cy2 < 0 ) cy2 = -cy2;
-	ppt->x = float2fixed(expand * max(cx1, cx2));
-	ppt->y = float2fixed(expand * max(cy1, cy2));
-	return 0;
+	{ float exx = expand * max(cx1, cx2);
+	  float exy = expand * max(cy1, cy2);
+	  int code = set_float2fixed_vars(ppt->x, exx);
+
+	  if ( code < 0 )
+	    return code;
+	  return set_float2fixed_vars(ppt->y, exy);
+	}
 }
 
 /*
@@ -124,6 +152,7 @@ typedef partial_line _ss *pl_ptr;
   ((pp)->x = (p).x, (pp)->y = (p).y)
 
 /* Other forward declarations */
+private bool width_is_thin(P1(pl_ptr));
 private void near adjust_stroke(P3(pl_ptr, const gs_imager_state *, bool));
 private int near line_join_points(P4(const gx_line_params *pgs_lp,
 				     pl_ptr plp, pl_ptr nplp,
@@ -146,12 +175,12 @@ gx_default_stroke_path(gx_device *dev, const gs_imager_state *pis,
 
 /* Fill a partial stroked path.  Free variables: */
 /* to_path, stroke_path_body, fill_params, always_thin, pis, dev, pdevc, */
-/* pcpath, code, ppath, exit(label). */
+/* code, ppath, exit(label). */
 #define fill_stroke_path(thin)\
 if(to_path==&stroke_path_body && !gx_path_is_void_inline(&stroke_path_body))\
 { fill_params.adjust.x = stroke_adjustment(thin, pis, x);\
   fill_params.adjust.y = stroke_adjustment(thin, pis, y);\
-  code = gx_fill_path_only(to_path, dev, pis, &fill_params, pdevc, pcpath);\
+  code = gx_fill_path_only(to_path, dev, pis, &fill_params, pdevc, NULL);\
   gx_path_release(to_path);\
   if ( code < 0 ) goto exit;\
   gx_path_init(to_path, ppath->memory);\
@@ -191,7 +220,6 @@ gx_stroke_path_only(gx_path *ppath, gx_path *to_path, gx_device *pdev,
 	gs_fixed_rect ibox, cbox;
 	gx_device_clip cdev;
 	gx_device *dev = pdev;
-	gx_device *save_dev = dev;
 	int code = 0;
 	gx_fill_params fill_params;
 	const gx_line_params *pgs_lp = gs_currentlineparams_inline(pis);
@@ -209,21 +237,27 @@ gx_stroke_path_only(gx_path *ppath, gx_path *to_path, gx_device *pdev,
 	int uniform;
 	bool reflected;
 	orientation orient =
-	  (/*is_fzero2(xy, yx) ?
+	  (
+#ifdef OPTIMIZE_ORIENTATION
+	   is_fzero2(xy, yx) ?
 		(uniform = (xx == yy ? 1 : xx == -yy ? -1 : 0),
 		 reflected = (uniform ? uniform < 0 : (xx < 0) != (yy < 0)),
 		 orient_portrait) :
 	   is_fzero2(xx, yy) ?
 		(uniform = (xy == yx ? -1 : xy == -yx ? 1 : 0),
 		 reflected = (uniform ? uniform < 0 : (xy < 0) == (yx < 0)),
-		 orient_landscape) :*/
+		 orient_landscape) :
+#endif
 	   (uniform = 0,
 	    reflected = xy * yx > xx * yy,
 	    orient_other));
+	segment_notes not_first =
+	  (!is_fzero(pis->line_params.dot_length) ? sn_not_first : sn_none);
 	float line_width = pgs_lp->half_width;  /* (*half* the line width) */
 	bool always_thin;
 	double line_width_and_scale, device_line_width_scale;
-	const segment *pseg;
+	double device_dot_length = pgs_lp->dot_length * fixed_1;
+	const subpath *psub;
 
 #ifdef DEBUG
 if ( gs_debug_c('o') )
@@ -245,27 +279,48 @@ if ( gs_debug_c('o') )
 	gx_path_bbox(ppath, &ibox);
 	/* Expand the path bounding box by the scaled line width. */
 	{ gs_fixed_point expansion;
-	  gx_stroke_expansion(pis, &expansion);
-	  expansion.x += pis->fill_adjust.x;
-	  expansion.y += pis->fill_adjust.y;
-	  ibox.p.x -= expansion.x;
-	  ibox.p.y -= expansion.y;
-	  ibox.q.x += expansion.x;
-	  ibox.q.y += expansion.y;
+
+	  if ( gx_stroke_path_expansion(pis, ppath, &expansion) < 0 )
+	    { /* The expansion is so large it caused a limitcheck. */
+	      ibox.p.x = ibox.p.y = min_fixed;
+	      ibox.q.x = ibox.q.y = max_fixed;
+	    }
+	  else
+	    { expansion.x += pis->fill_adjust.x;
+	      expansion.y += pis->fill_adjust.y;
+	      /*
+	       * It's theoretically possible for the following computations to
+	       * overflow, so we need to check for this.
+	       */
+	      ibox.p.x = (ibox.p.x < min_fixed + expansion.x ? min_fixed :
+			  ibox.p.x - expansion.x);
+	      ibox.p.y = (ibox.p.y < min_fixed + expansion.y ? min_fixed :
+			  ibox.p.y - expansion.y);
+	      ibox.q.x = (ibox.q.x > max_fixed - expansion.x ? max_fixed :
+			  ibox.q.x + expansion.x);
+	      ibox.q.y = (ibox.q.y > max_fixed - expansion.y ? max_fixed :
+			  ibox.q.y + expansion.x);
+	    }
 	}
 	/* Check the expanded bounding box against the clipping regions. */
-	if ( pcpath != NULL &&
-	     (gx_cpath_inner_box(pcpath, &cbox), !rect_within(ibox, cbox))
-	   )
-	   {	/* Intersect the path box and the clip bounding box. */
+	if ( pcpath )
+	  gx_cpath_inner_box(pcpath, &cbox);
+	else
+	  (*dev_proc(dev, get_clipping_box))(dev, &cbox);
+	if ( !rect_within(ibox, cbox) )
+	  {	/* Intersect the path box and the clip bounding box. */
 		/* If the intersection is empty, this call is a no-op. */
 		gs_fixed_rect bbox;
 
-		gx_cpath_outer_box(pcpath, &bbox);
-		if_debug4('f', "   outer_box=(%g,%g),(%g,%g)\n",
-			  fixed2float(bbox.p.x), fixed2float(bbox.p.y),
-			  fixed2float(bbox.q.x), fixed2float(bbox.q.y));
-		rect_intersect(ibox, bbox);
+		if ( pcpath )
+		  { gx_cpath_outer_box(pcpath, &bbox);
+		    if_debug4('f', "   outer_box=(%g,%g),(%g,%g)\n",
+			      fixed2float(bbox.p.x), fixed2float(bbox.p.y),
+			      fixed2float(bbox.q.x), fixed2float(bbox.q.y));
+		    rect_intersect(ibox, bbox);
+		  }
+		else
+		  rect_intersect(ibox, cbox);
 		if ( ibox.p.x >= ibox.q.x || ibox.p.y >= ibox.q.y )
 		{	/* Intersection of boxes is empty! */
 			return 0;
@@ -277,17 +332,24 @@ if ( gs_debug_c('o') )
 		 * recompute its bbox and make the tests again,
 		 * but we don't bother right now.
 		 *
-		 * Set up a clipping device.
+		 * If there is a clipping path, set up a clipping device.
 		 */
-		dev = (gx_device *)&cdev;
-		gx_make_clip_device(&cdev, &cdev, &pcpath->list);
-		cdev.target = save_dev;
-		cdev.max_fill_band = save_dev->max_fill_band;
-		(*dev_proc(dev, open_device))(dev);
-	   }
+		if ( pcpath )
+		  { gx_make_clip_device(&cdev, &cdev, &pcpath->list);
+		    cdev.target = dev;
+		    cdev.max_fill_band = dev->max_fill_band;
+		    dev = (gx_device *)&cdev;
+		    (*dev_proc(dev, open_device))(dev);
+		  }
+	  }
 	fill_params.rule = gx_rule_winding_number;
 	fill_params.flatness = pis->flatness;
-	fill_params.fill_zero_width = true;
+#ifdef USE_FILL_ADJUSTMENT
+	fill_params.fill_zero_width =
+	  (pis->fill_adjust.x | pis->fill_adjust.y) != 0;
+#else
+	fill_params.fill_zero_width = false;
+#endif
 	if ( line_width < 0 )
 	  line_width = -line_width;
 	line_width_and_scale = line_width * (double)int2fixed(1);
@@ -324,6 +386,22 @@ sat:		    if ( xa < 0 ) xa = -xa;
 	   }
 	if_debug5('o', "[o]ctm=(%g,%g,%g,%g) thin=%d\n",
 		  xx, xy, yx, yy, always_thin);
+	if ( device_dot_length != 0 )
+	  { /*
+	     * Compute the dot length in device space.  We can't do this
+	     * quite right for non-uniform coordinate systems; too bad.
+	     */
+	    gs_matrix mat;
+	    const gs_matrix *pmat;
+
+	    if ( pgs_lp->dot_length_absolute )
+	      { gs_deviceinitialmatrix(pdev, &mat);
+	        pmat = &mat;
+	      }
+	    else
+	      pmat = (const gs_matrix *)&pis->ctm;
+	    device_dot_length *= fabs(pmat->xy) + fabs(pmat->yy);
+	  }
 	/* Start by flattening the path.  We should do this on-the-fly.... */
 	if ( !ppath->curve_count )	/* don't need to flatten */
 	   {	if ( !ppath->first_subpath )
@@ -331,7 +409,8 @@ sat:		    if ( xa < 0 ) xa = -xa;
 		spath = ppath;
 	   }
 	else
-	   {	if ( (code = gx_path_flatten(ppath, &fpath, params->flatness)) < 0
+	   {	if ( (code = gx_path_flatten_accurate(ppath, &fpath,
+			params->flatness, pis->accurate_curves)) < 0
 		   )
 		  return code;
 		spath = &fpath;
@@ -347,18 +426,22 @@ sat:		    if ( xa < 0 ) xa = -xa;
 		to_path = &stroke_path_body;
 		gx_path_init(to_path, ppath->memory);
 	  }
-	for ( pseg = (const segment *)spath->first_subpath; pseg != 0; )
-	  { int first = 0;
-	    int index = 0;
+	for ( psub = spath->first_subpath; psub != 0; )
+	  { int index = 0;
+	    const segment *pseg = (const segment *)psub;
 	    fixed x = pseg->pt.x;
 	    fixed y = pseg->pt.y;
 	    bool is_closed = ((const subpath *)pseg)->is_closed;
 	    partial_line pl, pl_prev, pl_first;
-	    while ( (pseg = pseg->next) != 0 && pseg->type != s_start )
+
+	    while ( (pseg = pseg->next) != 0 &&
+		    pseg->type != s_start
+		  )
 	      {	/* Compute the width parameters in device space. */
 		/* We work with unscaled values, for speed. */
 		fixed sx = pseg->pt.x, udx = sx - x;
 		fixed sy = pseg->pt.y, udy = sy - y;
+
 		pl.o.p.x = x, pl.o.p.y = y;
 d:		pl.e.p.x = sx, pl.e.p.y = sy;
 		if ( !(udx | udy) )	/* degenerate */
@@ -375,48 +458,71 @@ d:		pl.e.p.x = sx, pl.e.p.y = sy;
 			  )
 		      { sx = pseg->pt.x, udx = sx - x;
 			sy = pseg->pt.y, udy = sy - y;
-			if ( (udx | udy) )
+			if ( udx | udy )
 			  goto d;
 		      }
 		    /*
 		     * The entire subpath is degenerate, but it includes
-		     * more than one point.  If we are using round caps,
-		     * draw the cap, otherwise do nothing.
+		     * more than one point.  If we are using round caps or
+		     * the dot length is non-zero, draw the caps, otherwise
+		     * do nothing.
 		     */
-		    if ( pgs_lp->cap == gs_cap_round )
-		      { /*
-			 * Set up cdelta; don't bother with efficiency.
-			 * This is the same computation as for the slow case
-			 * below, except that we arbitrarily choose the
-			 * direction so that dpt.x = fixed_1, dpt.y = 0.
-			 */
-			double dptx = line_width_and_scale;
-			pl.e.cdelta.x = (fixed)(dptx * xx);
-			pl.e.cdelta.y = (fixed)(dptx * xy);
-			if ( !reflected )
-			  dptx = -dptx;
-			pl.width.x = -(fixed)(dptx * yx);
-			pl.width.y = -(fixed)(dptx * yy);
-			pl.thin = false;
-			compute_caps(&pl);
-			/* To produce a complete dot, we need two */
-			/* round caps. */
-			code = gx_path_add_point(to_path, pl.e.co.x,
-						 pl.e.co.y);
-			if ( code < 0 )
-			  goto exit;
-			code = add_round_cap(to_path, &pl.e);
-			if ( code < 0 )
-			  goto exit;
-			code = add_round_cap(to_path, &pl.o);
-			if ( code < 0 )
-			  goto exit;
-			fill_stroke_path(false);
-		      }
-		    break;
+		    if ( !(pgs_lp->cap == gs_cap_round ||
+			   pgs_lp->dot_length != 0)
+		       )
+		      break;
+		    /*
+		     * Orient the dot according to the previous segment if
+		     * any, or else the next segment if any, or else a
+		     * vertical line.
+		     */
+		    { const segment *end = psub->prev;
+		      if ( end != 0 && (end->pt.x != x || end->pt.y != y) )
+			sx = end->pt.x, sy = end->pt.y;
+		      else if ( pseg != 0 &&
+				(pseg->pt.x != x || pseg->pt.y != y)
+			      )
+			sx = pseg->pt.x, sy = pseg->pt.y;
+		    }
+		    /*
+		     * Compute the properly oriented dot length, and then
+		     * draw the dot like a very short line.
+		     */
+		    udx = sx - x, udy = sy - y;
+		    if ( (udx | udy) == 0 )
+		      udy = fixed_1;
+		    { double scale = device_dot_length /
+			hypot((double)udx, (double)udy);
+
+		      /*
+		       * If we're using butt caps, make sure the "line" is
+		       * long enough to show up.
+		       */
+		      if ( pgs_lp->cap == gs_cap_butt )
+			{ fixed dmax = max(any_abs(udx), any_abs(udy));
+			  if ( dmax * scale < fixed_1 )
+			    scale = (float)fixed_1 / dmax;
+			}
+		      udx = (fixed)(udx * scale);
+		      udy = (fixed)(udy * scale);
+		      if ( (udx | udy) == 0 )
+			udy = fixed_epsilon;
+		      sx = x + udx;
+		      sy = y + udy;
+		    }
+		    /*
+		     * Back up 1 segment to keep the bookkeeping straight.
+		     */
+		    pseg = (pseg != 0 ? pseg->prev : psub->last);
+		    goto d;
 		  }
-		if ( !always_thin )
-		   {	if ( uniform != 0 )
+		if ( always_thin )
+		  {	pl.e.cdelta.x = pl.e.cdelta.y = 0;
+			pl.width.x = pl.width.y = 0;
+			pl.thin = true;
+		  }
+		else
+		  {	if ( uniform != 0 )
 			   {	/* We can save a lot of work in this case. */
 				/* We know orient != orient_other. */
 				float dpx = udx, dpy = udy;
@@ -465,43 +571,59 @@ d:		pl.e.p.x = sx, pl.e.p.y = sy;
 				if ( orient != orient_portrait )
 				  pl.width.x -= (fixed)(dpt.x * yx),
 				  pl.width.y += (fixed)(dpt.y * xy);
-				pl.thin =
-				  any_abs(pl.width.x) + any_abs(pl.width.y) <
-				    float2fixed(0.75);
+				pl.thin = width_is_thin(&pl);
 			   }
 			if ( !pl.thin )
 			{	adjust_stroke(&pl, pis, false);
 				compute_caps(&pl);
 			}
-		   }
-		else			/* always_thin */
-			pl.e.cdelta.x = pl.e.cdelta.y = 0,
-			pl.width.x = pl.width.y = 0,
-			pl.thin = true;
-		if ( first++ == 0 ) pl_first = pl;
+		  }
 		if ( index++ )
-		{	code = (*line_proc)(to_path,
-					    (is_closed ? 1 : index - 2),
-					    &pl_prev, &pl, pdevc, dev, pis,
-					    params, &cbox);
+		  {	int first;
+			pl_ptr lptr;
+
+			if ( pgs_lp->join == gs_join_none &&
+			     !(pseg->notes & not_first)
+			   )
+			  { /* Fake the end of a subpath so we get */
+			    /* caps instead of joins. */
+			    first = 0;
+			    lptr = 0;
+			    index = 1;
+			  }
+			else
+			  { first = (is_closed ? 1 : index - 2);
+			    lptr = &pl;
+			  }
+			code = (*line_proc)(to_path, first, &pl_prev, lptr,
+					    pdevc, dev, pis, params, &cbox);
 			if ( code < 0 )
 			  goto exit;
 			fill_stroke_path(always_thin);
-		}
+		  }
+		else
+		  pl_first = pl;
 		pl_prev = pl;
 		x = sx, y = sy;
 	      }
 	    if ( index )
 	      {	/* If closed, join back to start, else cap. */
 		/* For some reason, the Borland compiler requires the cast */
-		/* in the following line. */
-		pl_ptr lptr = (is_closed ? (pl_ptr)&pl_first : (pl_ptr)0);
+		/* in the following statement. */
+		pl_ptr lptr =
+		  (!is_closed ||
+		   (pgs_lp->join == gs_join_none &&
+		    !( (pseg == 0 ? (const segment *)spath->first_subpath :
+			pseg)->notes & not_first )) ?
+		   (pl_ptr)0 : (pl_ptr)&pl_first);
+
 		code = (*line_proc)(to_path, index - 1, &pl_prev, lptr,
 				    pdevc, dev, pis, params, &cbox);
 		if ( code < 0 )
 		  goto exit;
 		fill_stroke_path(always_thin);
 	      }
+	    psub = (const subpath *)pseg;
 	 }
 exit:	if ( to_path == &stroke_path_body )
 	  gx_path_release(to_path);	/* (only needed if error) */
@@ -514,6 +636,46 @@ exf:	if ( ppath->curve_count )
 
 /* ------ Internal routines ------ */
 
+/*
+ * Test whether a line is thin, i.e., whether the half-width, measured
+ * perpendicular to the line in device space, is less than 0.5 pixel.
+ * Unfortunately, the width values we computed are perpendicular to the
+ * line in *user* space, so we may have to do some extra work.
+ */
+private bool
+width_is_thin(pl_ptr plp)
+{	fixed dx, dy, wx = plp->width.x, wy = plp->width.y;
+
+	/* If the line is horizontal or vertical, things are easy. */
+	if ( (dy = plp->e.p.y - plp->o.p.y) == 0 )
+	  return any_abs(wy) < fixed_half;
+	if ( (dx = plp->e.p.x - plp->o.p.x) == 0 )
+	  return any_abs(wx) < fixed_half;
+
+	/*
+	 * If both horizontal and vertical widths are less than
+	 * 0.5, the line is thin.
+	 */
+	if ( any_abs(wx) < fixed_half && any_abs(wy) < fixed_half )
+	  return true;
+
+	/*
+	 * We have to do this the hard way, by actually computing the
+	 * perpendicular distance.  The distance from the point (U,V)
+	 * from a line from (0,0) to (C,D) is
+	 *	abs(C*V - D*U) / sqrt(C^2 + D^2)
+	 * In this case, (U,V) is plp->width, and (C,D) is (dx,dy).
+	 */
+	{ double C = dx, D = dy;
+	  double num = C * wy - D * wx;
+	  double denom = hypot(C, D);
+
+	  /* both num and denom are scaled by fixed_scale^2, */
+	  /* so we don't need to do any de-scaling for the test. */
+	  return fabs(num) < denom * 0.5;
+	}
+}
+
 /* Adjust the endpoints and width of a stroke segment */
 /* to achieve more uniform rendering. */
 /* Only o.p, e.p, e.cdelta, and width have been set. */
@@ -633,6 +795,7 @@ set_thin_widths(register pl_ptr plp)
 }
 
 /* Draw a line on the device. */
+/* Treat no join the same as a bevel join. */
 private int near
 stroke_fill(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
   const gx_device_color *pdevc, gx_device *dev, const gs_imager_state *pis,
@@ -670,7 +833,8 @@ stroke_fill(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
 	{ const gx_line_params *pgs_lp = gs_currentlineparams_inline(pis);
 	  gs_line_cap cap = pgs_lp->cap;
 	  gs_line_join join = pgs_lp->join;
-	  if ( !plp->thin
+
+	  if ( !plp->thin && (nplp == 0 || !nplp->thin)
 	      && ((first != 0 && nplp != 0) || cap == gs_cap_butt
 		  || cap == gs_cap_square)
 	      && (join == gs_join_bevel || join == gs_join_miter ||
@@ -680,7 +844,6 @@ stroke_fill(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
 	     )
 	    { gs_fixed_point points[6];
 	      int npoints, code;
-	      gs_fixed_point *bevel = points + 2;
 
 	      npoints = cap_points((first == 0 ? cap : gs_cap_butt),
 				   &plp->o, points);
@@ -690,7 +853,7 @@ stroke_fill(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
 		code = line_join_points(pgs_lp, plp, nplp, points + npoints);
 	      if ( code < 0 )
 		return code;
-	      if ( nplp != 0 && join != gs_join_none )
+	      if ( nplp != 0 )
 		{ if ( join == gs_join_miter )
 		    { /* Make sure we have a bevel and not a miter. */
 		      if ( !(points[2].x == plp->e.co.x &&
@@ -700,21 +863,23 @@ stroke_fill(gx_path *ppath, int first, register pl_ptr plp, pl_ptr nplp,
 			 )
 			goto fill;
 		    }
-		  /* Identify which 3 points define the bevel triangle. */
-		  if ( points[3].x == nplp->o.p.x &&
-		       points[3].y == nplp->o.p.y
-		     )
-		    ++bevel;
-		  /* Fill the bevel. */
-		  code = (*dev_proc(dev, fill_triangle))(dev,
+		  { gs_fixed_point *bevel = points + 2;
+		    /* Identify which 3 points define the bevel triangle. */
+		    if ( points[3].x == nplp->o.p.x &&
+			 points[3].y == nplp->o.p.y
+		       )
+		      ++bevel;
+		    /* Fill the bevel. */
+		    code = (*dev_proc(dev, fill_triangle))(dev,
 						bevel->x, bevel->y,
 						bevel[1].x - bevel->x,
 						bevel[1].y - bevel->y,
 						bevel[2].x - bevel->x,
 						bevel[2].y - bevel->y,
 						pdevc, pis->log_op);
-		  if ( code < 0 )
-		    return code;
+		    if ( code < 0 )
+		      return code;
+		  }
 		}
 	      /* Fill the body of the stroke. */
 	      return (*dev_proc(dev, fill_parallelogram))(dev,
@@ -820,10 +985,11 @@ add_points(gx_path *ppath, const gs_fixed_point _ss *points, int npoints,
 /* ---------------- Join computation ---------------- */
 
 /* Compute the points for a bevel, miter, or triangle join. */
+/* Treat no join the same as a bevel join. */
 private int near
 line_join_points(const gx_line_params *pgs_lp, pl_ptr plp, pl_ptr nplp,
   gs_fixed_point _ss *join_points)
-{		int num_points;
+{	gs_line_join join =  pgs_lp->join;
 #define jp1 join_points[0]
 #define np1 join_points[1]
 #define np2 join_points[2]
@@ -861,13 +1027,6 @@ line_join_points(const gx_line_params *pgs_lp, pl_ptr plp, pl_ptr nplp,
 
 		/* Initialize for a bevel join. */
 		assign_point(&jp1, plp->e.co);
-
-		if ( pgs_lp->join == gs_join_none )
-		  { /* No join at all, things are simple. */
-		    assign_point(&join_points[1], plp->e.ce);
-		    return 2;
-		  }
-
 		assign_point(&jp2, plp->e.ce);
 
 		/* Because of stroke adjustment, it is possible that */
@@ -888,7 +1047,7 @@ line_join_points(const gx_line_params *pgs_lp, pl_ptr plp, pl_ptr nplp,
 		if_debug1('O', "[o]use %s\n", (ccw ? "co (ccw)" : "ce (cw)"));
 
 		/* Handle triangular joins now. */
-		if ( pgs_lp->join == gs_join_triangle )
+		if ( join == gs_join_triangle )
 		  {	fixed tpx = outp->x - nplp->o.p.x + np->x;
 			fixed tpy = outp->y - nplp->o.p.y + np->y;
 			assign_point(&jpx, jp2);
@@ -902,15 +1061,13 @@ line_join_points(const gx_line_params *pgs_lp, pl_ptr plp, pl_ptr nplp,
 				assign_point(&np2, np1);
 				np1.x = tpx, np1.y = tpy;
 			  }
-			num_points = 5;
-			goto jadd;
+			return 5;
 		  }
-		num_points = 4;
 
 		/* Don't bother with the miter check if the two */
 		/* points to be joined are very close together, */
 		/* namely, in the same square half-pixel. */
-		if ( pgs_lp->join == gs_join_miter &&
+		if ( join == gs_join_miter &&
 		     !(fixed2long(outp->x << 1) == fixed2long(np->x << 1) &&
 		       fixed2long(outp->y << 1) == fixed2long(np->y << 1))
 		   )
@@ -977,7 +1134,7 @@ if ( gs_debug_c('O') )
 			  assign_point(outp, mpt);
 		    }
 		   }
-jadd:		return num_points;
+		return 4;
 }
 /* ---------------- Cap computations ---------------- */
 
