@@ -1,24 +1,19 @@
-/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1995, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* iscan.c */
@@ -193,11 +188,10 @@ private CLEAR_MARKS_PROC(scanner_clear_marks) {
 }
 private ENUM_PTRS_BEGIN(scanner_enum_ptrs) return 0;
 	case 0:
-		*pep = 0;
 		if ( ssptr->s_scan_type == scanning_none ||
 		     !ssptr->s_da.is_dynamic
 		   )
-		  break;
+		  ENUM_RETURN(0);
 		ssptr->s_da.str.data = ssptr->s_da.base;
 		ssptr->s_da.str.size = da_size(&ssptr->s_da);
 		ENUM_RETURN_STRING_PTR(scanner_state, s_da.str);
@@ -709,6 +703,7 @@ cend:		      /* Check for externally processed comments. */
 		  }
 		da.base = comment_line;
 		da.is_dynamic = false;
+	}
 		/* Enter here to continue scanning a comment. */
 		/* daptr must be set. */
 cont_comment:	for ( ; ; )
@@ -744,11 +739,9 @@ end_comment:		retcode = scan_comment(comment_line, daptr, true);
 			  goto sret;
 			goto top;
 		      }
-		    break;
 		  }
 #undef comment_line
 		/*NOTREACHED*/
-	}
 	case EOFC:
 		if ( pstack != 0 )
 		  sreturn(e_syntaxerror)
@@ -837,17 +830,18 @@ nr:		if ( retcode == 1 && decoder[newptr[-1]] == ctype_space )
 		name_type = 0;
 		try_number = false;
 do_name:
-	{	/* Try to scan entirely within the stream buffer. */
+		/* Try to scan entirely within the stream buffer. */
 		/* We stop 1 character early, so we don't switch buffers */
 		/* looking ahead if the name is terminated by \r\n. */
-		const byte *endp1 = endptr - 1;
 		da.base = (byte *)sptr;
 		da.is_dynamic = false;
-		do
-		{	if ( sptr >= endp1 )	/* stop 1 early! */
-				goto dyn_name;
+		{ const byte *endp1 = endptr - 1;
+		  do
+		    { if ( sptr >= endp1 )	/* stop 1 early! */
+		        goto dyn_name;
+		    }
+		  while ( decoder[*++sptr] <= max_name_ctype );	/* digit or name */
 		}
-		while ( decoder[*++sptr] <= max_name_ctype );	/* digit or name */
 		/* Name ended within the buffer. */
 		daptr = (byte *)sptr;
 		c = *sptr;
@@ -958,7 +952,6 @@ nx:		switch ( decoder[c] )
 		{	retcode = name_ref(da.base, (uint)(daptr - da.base),
 					   myref, 1);
 		}
-	}
 		/* Done scanning.  Check for preceding /'s. */
 		if ( retcode < 0 )
 		  {	if ( retcode != e_VMerror )
