@@ -1,24 +1,19 @@
-/* Copyright (C) 1991, 1995 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1991, 1995, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* time_.h */
@@ -33,8 +28,8 @@
 #include "gconfig_.h"
 
 /* Some System V environments don't include sys/time.h. */
-/* The SYSTIME_H switch in gconfig_.h reflects this. */
-#ifdef SYSTIME_H
+/* The HAVE_SYS_TIME_H switch in gconfig_.h reflects this. */
+#ifdef HAVE_SYS_TIME_H
 #  include <sys/time.h>
 #  if defined(M_UNIX) || defined(_IBMR2)	/* SCO and AIX need both time.h and sys/time.h! */
 #    include <time.h>
@@ -67,7 +62,7 @@ struct timezone {
 #endif
 
 /* Some System V environments, and Posix environments, need <sys/times.h>. */
-#if defined(SYSV) || defined(SVR4)
+#ifdef HAVE_SYS_TIMES_H
 #  include <sys/times.h>
 #  define use_times_for_usertime 1
 		/* Posix 1003.1b-1993 section 4.8.1.5 says that
