@@ -1,24 +1,19 @@
-/* Copyright (C) 1989, 1996 Aladdin Enterprises.  All rights reserved.
+/* Copyright (C) 1989, 1996, 1997 Aladdin Enterprises.  All rights reserved.
   
-  This file is part of GNU Ghostscript.
+  This file is part of Aladdin Ghostscript.
   
-  GNU Ghostscript is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
-  anyone for the consequences of using it or for whether it serves any
-  particular purpose or works at all, unless he says so in writing.  Refer to
-  the GNU General Public License for full details.
+  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
+  or distributor accepts any responsibility for the consequences of using it,
+  or for whether it serves any particular purpose or works at all, unless he
+  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
+  License (the "License") for full details.
   
-  Everyone is granted permission to copy, modify and redistribute GNU
-  Ghostscript, but only under the conditions described in the GNU General
-  Public License.  A copy of this license is supposed to have been given to
-  you along with GNU Ghostscript so you can know your rights and
-  responsibilities.  It should be in a file named COPYING.  Among other
-  things, the copyright notice and this notice must be preserved on all
-  copies.
-  
-  Aladdin Enterprises is not affiliated with the Free Software Foundation or
-  the GNU Project.  GNU Ghostscript, as distributed by Aladdin Enterprises,
-  does not depend on any other GNU software.
+  Every copy of Aladdin Ghostscript must include a copy of the License,
+  normally in a plain ASCII text file named PUBLIC.  The License grants you
+  the right to copy, modify and redistribute Aladdin Ghostscript, but only
+  under certain conditions described in the License.  Among other things, the
+  License requires that the copyright notice and this notice be preserved on
+  all copies.
 */
 
 /* zchar.c */
@@ -30,7 +25,7 @@
 #include "gxarith.h"
 #include "gxfixed.h"
 #include "gxmatrix.h"			/* for ifont.h */
-#include "gschar.h"
+#include "gxchar.h"			/* for stringwidth_flag */
 #include "gxdevice.h"			/* for gxfont.h */
 #include "gxfont.h"
 #include "gzpath.h"
@@ -327,6 +322,15 @@ END_OP_DEFS(0) }
 
 /* Most of these are exported for zchar2.c. */ 
 
+/* Convert a glyph to a ref. */
+private void
+glyph_ref(gs_glyph glyph, ref *gref)
+{	if ( glyph < gs_min_cid_glyph )
+	  name_index_ref(glyph, gref);
+	else
+	  make_int(gref, glyph - gs_min_cid_glyph);
+}
+
 /* Prepare to set up for a show operator. */
 /* Don't change any state yet. */
 int
@@ -362,6 +366,7 @@ op_show_finish_setup(gs_show_enum *penum, int npop, op_proc_p endproc /* end pro
 	make_int(&esodepth(ep), 0);	/* see gs_show_render case in */
 					/* op_show_continue_dispatch */
 	make_int(&esddepth(ep), 0);	/* ditto */
+	make_int(&esgslevel(ep), igs->level);
 	make_op_estack(&eseproc(ep), endproc);
 	make_istruct(ep, 0, penum);
 }
@@ -384,7 +389,7 @@ op_show_continue_dispatch(register os_ptr op, int code)
 	{	os_ptr save_osp = osp;
 		osp = op;
 		code = (*real_opproc(&seproc))(op);
-		op_show_free();
+		op_show_free(code);
 		if ( code < 0 )
 		{	osp = save_osp;
 			return code;
@@ -417,9 +422,10 @@ op_show_continue_dispatch(register os_ptr op, int code)
 		if ( pfont->FontType == ft_user_defined )
 		{	/* Type 3 font, prefer BuildGlyph. */
 			if ( level2_enabled &&
-			     !r_has_type(&pfdata->BuildGlyph, t_null)
+			     !r_has_type(&pfdata->BuildGlyph, t_null) &&
+			     glyph != gs_no_glyph
 			   )
-			{	name_index_ref(glyph, op);
+			{	glyph_ref(glyph, op);
 				esp[2] = pfdata->BuildGlyph;
 			}
 			else if ( r_has_type(&pfdata->BuildChar, t_null) )
@@ -429,7 +435,8 @@ op_show_continue_dispatch(register os_ptr op, int code)
 				/* through the Encoding */
 				ref gref;
 				const ref *pencoding = &pfdata->Encoding;
-				name_index_ref(glyph, &gref);
+
+				glyph_ref(glyph, &gref);
 				if ( !map_glyph_to_char(&gref, pencoding,
 							(ref *)op)
 				   )
@@ -457,7 +464,7 @@ op_show_continue_dispatch(register os_ptr op, int code)
 				esp[2] = pfdata->BuildChar;
 			}
 			else
-			{	name_index_ref(glyph, op);
+			{	glyph_ref(glyph, op);
 				esp[2] = pfdata->BuildGlyph;
 			}
 		}
@@ -469,11 +476,9 @@ op_show_continue_dispatch(register os_ptr op, int code)
 	}
 		return o_push_estack;
 	default:			/* error */
-err:		op_show_free();
-		if ( code < 0 )
-		  return code;
-		else
-		  return_error(e_invalidfont);
+err:		if ( code >= 0 )
+		  code = gs_note_error(e_invalidfont);
+		return op_show_free(code);
 	}
 }
 /* Reverse-map a glyph name to a character code for glyphshow. */
@@ -549,20 +554,54 @@ op_show_return_width(os_ptr op, uint npop, float *pwidth)
 	return o_pop_estack;
 }
 
-/* Discard the show record (after an error, or at the end). */
+/*
+ * Restore state after finishing, or unwinding from an error within, a show
+ * operation.  Note that we assume op == osp, and may reset osp.
+ */
 private int
-op_show_cleanup(os_ptr op)
+op_show_restore(bool for_error)
 {	register es_ptr ep = esp + snumpush;
 	gs_show_enum *penum = esenum(ep);
-	if ( r_is_struct(&esslot(ep)) )
-	  ifree_object(esslot(ep).value.pstruct, "free_show(stream)");
+	int saved_level = esgslevel(ep).value.intval;
+	int code = 0;
+
+	if ( for_error )
+	  { uint saved_count = esodepth(ep).value.intval;
+	    uint count = ref_stack_count(&o_stack);
+
+	    if ( count > saved_count ) /* if <, we're in trouble */
+	      ref_stack_pop(&o_stack, count - saved_count);
+	  }
+	if ( penum->stringwidth_flag == 1 )
+	  { /* stringwidth does an extra gsave */
+	    --saved_level;
+	  }
+	while ( igs->level > saved_level && code >= 0 )
+	  { if ( igs->saved == 0 || igs->saved->saved == 0 )
+	      { /*
+		 * Bad news: we got an error inside a save inside a
+		 * BuildChar or BuildGlyph.  Don't attempt to recover.
+		 */
+		code = gs_note_error(e_Fatal);
+	      }
+	    else
+	      code = gs_grestore(igs);
+	  }
 	gs_show_enum_release(penum, (gs_memory_t *)imemory);
-	return 0;
+	return code;
 }
-void
-op_show_free(void)
-{	esp -= snumpush;
-	op_show_cleanup(osp);
+/* Clean up after an error. */
+private int
+op_show_cleanup(os_ptr op)
+{	return op_show_restore(true);
+}
+/* Clean up after termination of a show operation. */
+int
+op_show_free(int code)
+{	int rcode;
+	esp -= snumpush;
+	rcode = op_show_restore(code < 0);
+	return (rcode < 0 ? rcode : code);
 }
 
 /* Get a FontBBox parameter from a font dictionary. */
